"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

var _typeof = require("@babel/runtime/helpers/typeof");

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _toConsumableArray2 = _interopRequireDefault(require("@babel/runtime/helpers/toConsumableArray"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _logger = require("../../logger");

var utils = _interopRequireWildcard(require("../../utils"));

var olmlib = _interopRequireWildcard(require("../olmlib"));

var _base = require("./base");

var _OlmDevice = require("../OlmDevice");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || _typeof(obj) !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _createForOfIteratorHelper(o, allowArrayLike) { var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"]; if (!it) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = it.call(o); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

// determine whether the key can be shared with invitees
function isRoomSharedHistory(room) {
  var visibilityEvent = room.currentState && room.currentState.getStateEvents("m.room.history_visibility", ""); // NOTE: if the room visibility is unset, it would normally default to
  // "world_readable".
  // (https://spec.matrix.org/unstable/client-server-api/#server-behaviour-5)
  // But we will be paranoid here, and treat it as a situation where the room
  // is not shared-history

  var visibility = visibilityEvent && visibilityEvent.getContent() && visibilityEvent.getContent().history_visibility;
  return ["world_readable", "shared"].includes(visibility);
}
/**
 * @private
 * @constructor
 *
 * @param {string} sessionId
 * @param {boolean} sharedHistory whether the session can be freely shared with
 *    other group members, according to the room history visibility settings
 *
 * @property {string} sessionId
 * @property {Number} useCount     number of times this session has been used
 * @property {Number} creationTime when the session was created (ms since the epoch)
 *
 * @property {object} sharedWithDevices
 *    devices with which we have shared the session key
 *        userId -> {deviceId -> msgindex}
 */


function OutboundSessionInfo(sessionId) {
  var sharedHistory = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
  this.sessionId = sessionId;
  this.useCount = 0;
  this.creationTime = new Date().getTime();
  this.sharedWithDevices = {};
  this.blockedDevicesNotified = {};
  this.sharedHistory = sharedHistory;
}
/**
 * Check if it's time to rotate the session
 *
 * @param {Number} rotationPeriodMsgs
 * @param {Number} rotationPeriodMs
 * @return {Boolean}
 */


OutboundSessionInfo.prototype.needsRotation = function (rotationPeriodMsgs, rotationPeriodMs) {
  var sessionLifetime = new Date().getTime() - this.creationTime;

  if (this.useCount >= rotationPeriodMsgs || sessionLifetime >= rotationPeriodMs) {
    _logger.logger.log("Rotating megolm session after " + this.useCount + " messages, " + sessionLifetime + "ms");

    return true;
  }

  return false;
};

OutboundSessionInfo.prototype.markSharedWithDevice = function (userId, deviceId, chainIndex) {
  if (!this.sharedWithDevices[userId]) {
    this.sharedWithDevices[userId] = {};
  }

  this.sharedWithDevices[userId][deviceId] = chainIndex;
};

OutboundSessionInfo.prototype.markNotifiedBlockedDevice = function (userId, deviceId) {
  if (!this.blockedDevicesNotified[userId]) {
    this.blockedDevicesNotified[userId] = {};
  }

  this.blockedDevicesNotified[userId][deviceId] = true;
};
/**
 * Determine if this session has been shared with devices which it shouldn't
 * have been.
 *
 * @param {Object} devicesInRoom userId -> {deviceId -> object}
 *   devices we should shared the session with.
 *
 * @return {Boolean} true if we have shared the session with devices which aren't
 * in devicesInRoom.
 */


OutboundSessionInfo.prototype.sharedWithTooManyDevices = function (devicesInRoom) {
  for (var userId in this.sharedWithDevices) {
    if (!this.sharedWithDevices.hasOwnProperty(userId)) {
      continue;
    }

    if (!devicesInRoom.hasOwnProperty(userId)) {
      _logger.logger.log("Starting new megolm session because we shared with " + userId);

      return true;
    }

    for (var deviceId in this.sharedWithDevices[userId]) {
      if (!this.sharedWithDevices[userId].hasOwnProperty(deviceId)) {
        continue;
      }

      if (!devicesInRoom[userId].hasOwnProperty(deviceId)) {
        _logger.logger.log("Starting new megolm session because we shared with " + userId + ":" + deviceId);

        return true;
      }
    }
  }
};
/**
 * Megolm encryption implementation
 *
 * @constructor
 * @extends {module:crypto/algorithms/EncryptionAlgorithm}
 *
 * @param {object} params parameters, as per
 *     {@link module:crypto/algorithms/EncryptionAlgorithm}
 */


function MegolmEncryption(params) {
  (0, utils.polyfillSuper)(this, _base.EncryptionAlgorithm, params); // the most recent attempt to set up a session. This is used to serialise
  // the session setups, so that we have a race-free view of which session we
  // are using, and which devices we have shared the keys with. It resolves
  // with an OutboundSessionInfo (or undefined, for the first message in the
  // room).

  this._setupPromise = Promise.resolve(); // Map of outbound sessions by sessions ID. Used if we need a particular
  // session (the session we're currently using to send is always obtained
  // using _setupPromise).

  this._outboundSessions = {}; // default rotation periods

  this._sessionRotationPeriodMsgs = 100;
  this._sessionRotationPeriodMs = 7 * 24 * 3600 * 1000;

  if (params.config.rotation_period_ms !== undefined) {
    this._sessionRotationPeriodMs = params.config.rotation_period_ms;
  }

  if (params.config.rotation_period_msgs !== undefined) {
    this._sessionRotationPeriodMsgs = params.config.rotation_period_msgs;
  }
}

utils.inherits(MegolmEncryption, _base.EncryptionAlgorithm);
/**
 * @private
 *
 * @param {module:models/room} room
 * @param {Object} devicesInRoom The devices in this room, indexed by user ID
 * @param {Object} blocked The devices that are blocked, indexed by user ID
 * @param {boolean} [singleOlmCreationPhase] Only perform one round of olm
 *     session creation
 *
 * @return {Promise} Promise which resolves to the
 *    OutboundSessionInfo when setup is complete.
 */

MegolmEncryption.prototype._ensureOutboundSession = /*#__PURE__*/function () {
  var _ref = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6(room, devicesInRoom, blocked, singleOlmCreationPhase) {
    var _this = this;

    var session, prepareSession, returnSession, prom;
    return _regenerator["default"].wrap(function _callee6$(_context6) {
      while (1) {
        switch (_context6.prev = _context6.next) {
          case 0:
            returnSession = function _returnSession() {
              return session;
            };

            // takes the previous OutboundSessionInfo, and considers whether to create
            // a new one. Also shares the key with any (new) devices in the room.
            // Updates `session` to hold the final OutboundSessionInfo.
            //
            // returns a promise which resolves once the keyshare is successful.
            prepareSession = /*#__PURE__*/function () {
              var _ref2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5(oldSession) {
                var sharedHistory, shareMap, _i, _Object$entries, _Object$entries$_i, userId, userDevices, _i2, _Object$entries2, _Object$entries2$_i, deviceId, deviceInfo, _key, key, payload, _yield$olmlib$getExis, _yield$olmlib$getExis2, devicesWithoutSession, olmSessions;

                return _regenerator["default"].wrap(function _callee5$(_context5) {
                  while (1) {
                    switch (_context5.prev = _context5.next) {
                      case 0:
                        session = oldSession;
                        sharedHistory = isRoomSharedHistory(room); // history visibility changed

                        if (session && sharedHistory !== session.sharedHistory) {
                          session = null;
                        } // need to make a brand new session?


                        if (session && session.needsRotation(_this._sessionRotationPeriodMsgs, _this._sessionRotationPeriodMs)) {
                          _logger.logger.log("Starting new megolm session because we need to rotate.");

                          session = null;
                        } // determine if we have shared with anyone we shouldn't have


                        if (session && session.sharedWithTooManyDevices(devicesInRoom)) {
                          session = null;
                        }

                        if (session) {
                          _context5.next = 12;
                          break;
                        }

                        _logger.logger.log("Starting new megolm session for room ".concat(_this._roomId));

                        _context5.next = 9;
                        return _this._prepareNewSession(sharedHistory);

                      case 9:
                        session = _context5.sent;

                        _logger.logger.log("Started new megolm session ".concat(session.sessionId, " ") + "for room ".concat(_this._roomId));

                        _this._outboundSessions[session.sessionId] = session;

                      case 12:
                        // now check if we need to share with any devices
                        shareMap = {};
                        _i = 0, _Object$entries = Object.entries(devicesInRoom);

                      case 14:
                        if (!(_i < _Object$entries.length)) {
                          _context5.next = 29;
                          break;
                        }

                        _Object$entries$_i = (0, _slicedToArray2["default"])(_Object$entries[_i], 2), userId = _Object$entries$_i[0], userDevices = _Object$entries$_i[1];
                        _i2 = 0, _Object$entries2 = Object.entries(userDevices);

                      case 17:
                        if (!(_i2 < _Object$entries2.length)) {
                          _context5.next = 26;
                          break;
                        }

                        _Object$entries2$_i = (0, _slicedToArray2["default"])(_Object$entries2[_i2], 2), deviceId = _Object$entries2$_i[0], deviceInfo = _Object$entries2$_i[1];
                        _key = deviceInfo.getIdentityKey();

                        if (!(_key == _this._olmDevice.deviceCurve25519Key)) {
                          _context5.next = 22;
                          break;
                        }

                        return _context5.abrupt("continue", 23);

                      case 22:
                        if (!session.sharedWithDevices[userId] || session.sharedWithDevices[userId][deviceId] === undefined) {
                          shareMap[userId] = shareMap[userId] || [];
                          shareMap[userId].push(deviceInfo);
                        }

                      case 23:
                        _i2++;
                        _context5.next = 17;
                        break;

                      case 26:
                        _i++;
                        _context5.next = 14;
                        break;

                      case 29:
                        key = _this._olmDevice.getOutboundGroupSessionKey(session.sessionId);
                        payload = {
                          type: "m.room_key",
                          content: {
                            "algorithm": olmlib.MEGOLM_ALGORITHM,
                            "room_id": _this._roomId,
                            "session_id": session.sessionId,
                            "session_key": key.key,
                            "chain_index": key.chain_index,
                            "org.matrix.msc3061.shared_history": sharedHistory
                          }
                        };
                        _context5.next = 33;
                        return olmlib.getExistingOlmSessions(_this._olmDevice, _this._baseApis, shareMap);

                      case 33:
                        _yield$olmlib$getExis = _context5.sent;
                        _yield$olmlib$getExis2 = (0, _slicedToArray2["default"])(_yield$olmlib$getExis, 2);
                        devicesWithoutSession = _yield$olmlib$getExis2[0];
                        olmSessions = _yield$olmlib$getExis2[1];
                        _context5.next = 39;
                        return Promise.all([(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee() {
                          return _regenerator["default"].wrap(function _callee$(_context) {
                            while (1) {
                              switch (_context.prev = _context.next) {
                                case 0:
                                  // share keys with devices that we already have a session for
                                  _logger.logger.debug("Sharing keys with existing Olm sessions in ".concat(_this._roomId));

                                  _context.next = 3;
                                  return _this._shareKeyWithOlmSessions(session, key, payload, olmSessions);

                                case 3:
                                  _logger.logger.debug("Shared keys with existing Olm sessions in ".concat(_this._roomId));

                                case 4:
                                case "end":
                                  return _context.stop();
                              }
                            }
                          }, _callee);
                        }))(), (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3() {
                          var errorDevices, start, failedServers;
                          return _regenerator["default"].wrap(function _callee3$(_context3) {
                            while (1) {
                              switch (_context3.prev = _context3.next) {
                                case 0:
                                  _logger.logger.debug("Sharing keys (start phase 1) with new Olm sessions in ".concat(_this._roomId));

                                  errorDevices = []; // meanwhile, establish olm sessions for devices that we don't
                                  // already have a session for, and share keys with them.  If
                                  // we're doing two phases of olm session creation, use a
                                  // shorter timeout when fetching one-time keys for the first
                                  // phase.

                                  // meanwhile, establish olm sessions for devices that we don't
                                  // already have a session for, and share keys with them.  If
                                  // we're doing two phases of olm session creation, use a
                                  // shorter timeout when fetching one-time keys for the first
                                  // phase.
                                  start = Date.now();
                                  failedServers = [];
                                  _context3.next = 6;
                                  return _this._shareKeyWithDevices(session, key, payload, devicesWithoutSession, errorDevices, singleOlmCreationPhase ? 10000 : 2000, failedServers);

                                case 6:
                                  _logger.logger.debug("Shared keys (end phase 1) with new Olm sessions in ".concat(_this._roomId));

                                  if (!(!singleOlmCreationPhase && Date.now() - start < 10000)) {
                                    _context3.next = 11;
                                    break;
                                  }

                                  // perform the second phase of olm session creation if requested,
                                  // and if the first phase didn't take too long
                                  (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
                                    var retryDevices, failedServerMap, _iterator, _step, server, failedDevices, _iterator2, _step2, _step2$value, _userId, _deviceInfo, userHS;

                                    return _regenerator["default"].wrap(function _callee2$(_context2) {
                                      while (1) {
                                        switch (_context2.prev = _context2.next) {
                                          case 0:
                                            // Retry sending keys to devices that we were unable to establish
                                            // an olm session for.  This time, we use a longer timeout, but we
                                            // do this in the background and don't block anything else while we
                                            // do this.  We only need to retry users from servers that didn't
                                            // respond the first time.
                                            retryDevices = {};
                                            failedServerMap = new Set();
                                            _iterator = _createForOfIteratorHelper(failedServers);

                                            try {
                                              for (_iterator.s(); !(_step = _iterator.n()).done;) {
                                                server = _step.value;
                                                failedServerMap.add(server);
                                              }
                                            } catch (err) {
                                              _iterator.e(err);
                                            } finally {
                                              _iterator.f();
                                            }

                                            failedDevices = [];
                                            _iterator2 = _createForOfIteratorHelper(errorDevices);

                                            try {
                                              for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
                                                _step2$value = _step2.value, _userId = _step2$value.userId, _deviceInfo = _step2$value.deviceInfo;
                                                userHS = _userId.slice(_userId.indexOf(":") + 1);

                                                if (failedServerMap.has(userHS)) {
                                                  retryDevices[_userId] = retryDevices[_userId] || [];

                                                  retryDevices[_userId].push(_deviceInfo);
                                                } else {
                                                  // if we aren't going to retry, then handle it
                                                  // as a failed device
                                                  failedDevices.push({
                                                    userId: _userId,
                                                    deviceInfo: _deviceInfo
                                                  });
                                                }
                                              }
                                            } catch (err) {
                                              _iterator2.e(err);
                                            } finally {
                                              _iterator2.f();
                                            }

                                            _logger.logger.debug("Sharing keys (start phase 2) with new Olm sessions in ".concat(_this._roomId));

                                            _context2.next = 10;
                                            return _this._shareKeyWithDevices(session, key, payload, retryDevices, failedDevices, 30000);

                                          case 10:
                                            _logger.logger.debug("Shared keys (end phase 2) with new Olm sessions in ".concat(_this._roomId));

                                            _context2.next = 13;
                                            return _this._notifyFailedOlmDevices(session, key, failedDevices);

                                          case 13:
                                          case "end":
                                            return _context2.stop();
                                        }
                                      }
                                    }, _callee2);
                                  }))();
                                  _context3.next = 13;
                                  break;

                                case 11:
                                  _context3.next = 13;
                                  return _this._notifyFailedOlmDevices(session, key, errorDevices);

                                case 13:
                                  _logger.logger.debug("Shared keys (all phases done) with new Olm sessions in ".concat(_this._roomId));

                                case 14:
                                case "end":
                                  return _context3.stop();
                              }
                            }
                          }, _callee3);
                        }))(), (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4() {
                          var blockedMap, blockedCount, _i3, _Object$entries3, _Object$entries3$_i, _userId2, userBlockedDevices, _i4, _Object$entries4, _Object$entries4$_i, _deviceId, device;

                          return _regenerator["default"].wrap(function _callee4$(_context4) {
                            while (1) {
                              switch (_context4.prev = _context4.next) {
                                case 0:
                                  _logger.logger.debug("Notifying blocked devices in ".concat(_this._roomId)); // also, notify blocked devices that they're blocked


                                  // also, notify blocked devices that they're blocked
                                  blockedMap = {};
                                  blockedCount = 0;

                                  for (_i3 = 0, _Object$entries3 = Object.entries(blocked); _i3 < _Object$entries3.length; _i3++) {
                                    _Object$entries3$_i = (0, _slicedToArray2["default"])(_Object$entries3[_i3], 2), _userId2 = _Object$entries3$_i[0], userBlockedDevices = _Object$entries3$_i[1];

                                    for (_i4 = 0, _Object$entries4 = Object.entries(userBlockedDevices); _i4 < _Object$entries4.length; _i4++) {
                                      _Object$entries4$_i = (0, _slicedToArray2["default"])(_Object$entries4[_i4], 2), _deviceId = _Object$entries4$_i[0], device = _Object$entries4$_i[1];

                                      if (!session.blockedDevicesNotified[_userId2] || session.blockedDevicesNotified[_userId2][_deviceId] === undefined) {
                                        blockedMap[_userId2] = blockedMap[_userId2] || {};
                                        blockedMap[_userId2][_deviceId] = {
                                          device: device
                                        };
                                        blockedCount++;
                                      }
                                    }
                                  }

                                  _context4.next = 6;
                                  return _this._notifyBlockedDevices(session, blockedMap);

                                case 6:
                                  _logger.logger.debug("Notified ".concat(blockedCount, " blocked devices in ").concat(_this._roomId));

                                case 7:
                                case "end":
                                  return _context4.stop();
                              }
                            }
                          }, _callee4);
                        }))()]);

                      case 39:
                      case "end":
                        return _context5.stop();
                    }
                  }
                }, _callee5);
              }));

              return function prepareSession(_x5) {
                return _ref2.apply(this, arguments);
              };
            }(); // helper which returns the session prepared by prepareSession


            // first wait for the previous share to complete
            prom = this._setupPromise.then(prepareSession); // Ensure any failures are logged for debugging

            prom["catch"](function (e) {
              _logger.logger.error("Failed to ensure outbound session in ".concat(_this._roomId), e);
            }); // _setupPromise resolves to `session` whether or not the share succeeds

            this._setupPromise = prom.then(returnSession, returnSession); // but we return a promise which only resolves if the share was successful.

            return _context6.abrupt("return", prom.then(returnSession));

          case 6:
          case "end":
            return _context6.stop();
        }
      }
    }, _callee6, this);
  }));

  return function (_x, _x2, _x3, _x4) {
    return _ref.apply(this, arguments);
  };
}();
/**
 * @private
 *
 * @param {boolean} sharedHistory
 *
 * @return {module:crypto/algorithms/megolm.OutboundSessionInfo} session
 */


MegolmEncryption.prototype._prepareNewSession = /*#__PURE__*/function () {
  var _ref7 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee7(sharedHistory) {
    var sessionId, key;
    return _regenerator["default"].wrap(function _callee7$(_context7) {
      while (1) {
        switch (_context7.prev = _context7.next) {
          case 0:
            sessionId = this._olmDevice.createOutboundGroupSession();
            key = this._olmDevice.getOutboundGroupSessionKey(sessionId);
            _context7.next = 4;
            return this._olmDevice.addInboundGroupSession(this._roomId, this._olmDevice.deviceCurve25519Key, [], sessionId, key.key, {
              ed25519: this._olmDevice.deviceEd25519Key
            }, false, {
              sharedHistory: sharedHistory
            });

          case 4:
            // don't wait for it to complete
            this._crypto.backupGroupSession(this._roomId, this._olmDevice.deviceCurve25519Key, [], sessionId, key.key);

            return _context7.abrupt("return", new OutboundSessionInfo(sessionId, sharedHistory));

          case 6:
          case "end":
            return _context7.stop();
        }
      }
    }, _callee7, this);
  }));

  return function (_x6) {
    return _ref7.apply(this, arguments);
  };
}();
/**
 * Determines what devices in devicesByUser don't have an olm session as given
 * in devicemap.
 *
 * @private
 *
 * @param {object} devicemap the devices that have olm sessions, as returned by
 *     olmlib.ensureOlmSessionsForDevices.
 * @param {object} devicesByUser a map of user IDs to array of deviceInfo
 * @param {array} [noOlmDevices] an array to fill with devices that don't have
 *     olm sessions
 *
 * @return {array} an array of devices that don't have olm sessions.  If
 *     noOlmDevices is specified, then noOlmDevices will be returned.
 */


MegolmEncryption.prototype._getDevicesWithoutSessions = function (devicemap, devicesByUser, noOlmDevices) {
  noOlmDevices = noOlmDevices || [];

  for (var _i5 = 0, _Object$entries5 = Object.entries(devicesByUser); _i5 < _Object$entries5.length; _i5++) {
    var _Object$entries5$_i = (0, _slicedToArray2["default"])(_Object$entries5[_i5], 2),
        userId = _Object$entries5$_i[0],
        devicesToShareWith = _Object$entries5$_i[1];

    var sessionResults = devicemap[userId];

    var _iterator3 = _createForOfIteratorHelper(devicesToShareWith),
        _step3;

    try {
      for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
        var deviceInfo = _step3.value;
        var deviceId = deviceInfo.deviceId;
        var sessionResult = sessionResults[deviceId];

        if (!sessionResult.sessionId) {
          // no session with this device, probably because there
          // were no one-time keys.
          noOlmDevices.push({
            userId: userId,
            deviceInfo: deviceInfo
          });
          delete sessionResults[deviceId]; // ensureOlmSessionsForUsers has already done the logging,
          // so just skip it.

          continue;
        }
      }
    } catch (err) {
      _iterator3.e(err);
    } finally {
      _iterator3.f();
    }
  }

  return noOlmDevices;
};
/**
 * Splits the user device map into multiple chunks to reduce the number of
 * devices we encrypt to per API call.
 *
 * @private
 *
 * @param {object} devicesByUser map from userid to list of devices
 *
 * @return {array<array<object>>} the blocked devices, split into chunks
 */


MegolmEncryption.prototype._splitDevices = function (devicesByUser) {
  var maxDevicesPerRequest = 20; // use an array where the slices of a content map gets stored

  var currentSlice = [];
  var mapSlices = [currentSlice];

  for (var _i6 = 0, _Object$entries6 = Object.entries(devicesByUser); _i6 < _Object$entries6.length; _i6++) {
    var _Object$entries6$_i = (0, _slicedToArray2["default"])(_Object$entries6[_i6], 2),
        userId = _Object$entries6$_i[0],
        userDevices = _Object$entries6$_i[1];

    for (var _i7 = 0, _Object$values = Object.values(userDevices); _i7 < _Object$values.length; _i7++) {
      var deviceInfo = _Object$values[_i7];
      currentSlice.push({
        userId: userId,
        deviceInfo: deviceInfo.device
      });
    } // We do this in the per-user loop as we prefer that all messages to the
    // same user end up in the same API call to make it easier for the
    // server (e.g. only have to send one EDU if a remote user, etc). This
    // does mean that if a user has many devices we may go over the desired
    // limit, but its not a hard limit so that is fine.


    if (currentSlice.length > maxDevicesPerRequest) {
      // the current slice is filled up. Start inserting into the next slice
      currentSlice = [];
      mapSlices.push(currentSlice);
    }
  }

  if (currentSlice.length === 0) {
    mapSlices.pop();
  }

  return mapSlices;
};
/**
 * @private
 *
 * @param {module:crypto/algorithms/megolm.OutboundSessionInfo} session
 *
 * @param {number} chainIndex current chain index
 *
 * @param {object<userId, deviceInfo>} userDeviceMap
 *   mapping from userId to deviceInfo
 *
 * @param {object} payload fields to include in the encrypted payload
 *
 * @return {Promise} Promise which resolves once the key sharing
 *     for the given userDeviceMap is generated and has been sent.
 */


MegolmEncryption.prototype._encryptAndSendKeysToDevices = function (session, chainIndex, userDeviceMap, payload) {
  var _this2 = this;

  var contentMap = {};
  var promises = [];

  for (var i = 0; i < userDeviceMap.length; i++) {
    var encryptedContent = {
      algorithm: olmlib.OLM_ALGORITHM,
      sender_key: this._olmDevice.deviceCurve25519Key,
      ciphertext: {}
    };
    var val = userDeviceMap[i];
    var userId = val.userId;
    var deviceInfo = val.deviceInfo;
    var deviceId = deviceInfo.deviceId;

    if (!contentMap[userId]) {
      contentMap[userId] = {};
    }

    contentMap[userId][deviceId] = encryptedContent;
    promises.push(olmlib.encryptMessageForDevice(encryptedContent.ciphertext, this._userId, this._deviceId, this._olmDevice, userId, deviceInfo, payload));
  }

  return Promise.all(promises).then(function () {
    // prune out any devices that encryptMessageForDevice could not encrypt for,
    // in which case it will have just not added anything to the ciphertext object.
    // There's no point sending messages to devices if we couldn't encrypt to them,
    // since that's effectively a blank message.
    for (var _i8 = 0, _Object$keys = Object.keys(contentMap); _i8 < _Object$keys.length; _i8++) {
      var _userId3 = _Object$keys[_i8];

      for (var _i9 = 0, _Object$keys2 = Object.keys(contentMap[_userId3]); _i9 < _Object$keys2.length; _i9++) {
        var _deviceId2 = _Object$keys2[_i9];

        if (Object.keys(contentMap[_userId3][_deviceId2].ciphertext).length === 0) {
          _logger.logger.log("No ciphertext for device " + _userId3 + ":" + _deviceId2 + ": pruning");

          delete contentMap[_userId3][_deviceId2];
        }
      } // No devices left for that user? Strip that too.


      if (Object.keys(contentMap[_userId3]).length === 0) {
        _logger.logger.log("Pruned all devices for user " + _userId3);

        delete contentMap[_userId3];
      }
    } // Is there anything left?


    if (Object.keys(contentMap).length === 0) {
      _logger.logger.log("No users left to send to: aborting");

      return;
    }

    return _this2._baseApis.sendToDevice("m.room.encrypted", contentMap).then(function () {
      // store that we successfully uploaded the keys of the current slice
      for (var _i10 = 0, _Object$keys3 = Object.keys(contentMap); _i10 < _Object$keys3.length; _i10++) {
        var _userId4 = _Object$keys3[_i10];

        for (var _i11 = 0, _Object$keys4 = Object.keys(contentMap[_userId4]); _i11 < _Object$keys4.length; _i11++) {
          var _deviceId3 = _Object$keys4[_i11];
          session.markSharedWithDevice(_userId4, _deviceId3, chainIndex);
        }
      }
    });
  });
};
/**
 * @private
 *
 * @param {module:crypto/algorithms/megolm.OutboundSessionInfo} session
 *
 * @param {array<object>} userDeviceMap list of blocked devices to notify
 *
 * @param {object} payload fields to include in the notification payload
 *
 * @return {Promise} Promise which resolves once the notifications
 *     for the given userDeviceMap is generated and has been sent.
 */


MegolmEncryption.prototype._sendBlockedNotificationsToDevices = /*#__PURE__*/function () {
  var _ref8 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee8(session, userDeviceMap, payload) {
    var contentMap, _iterator4, _step4, val, _userId5, blockedInfo, deviceInfo, _deviceId4, message, _i12, _Object$keys5, userId, _i13, _Object$keys6, deviceId;

    return _regenerator["default"].wrap(function _callee8$(_context8) {
      while (1) {
        switch (_context8.prev = _context8.next) {
          case 0:
            contentMap = {};
            _iterator4 = _createForOfIteratorHelper(userDeviceMap);

            try {
              for (_iterator4.s(); !(_step4 = _iterator4.n()).done;) {
                val = _step4.value;
                _userId5 = val.userId;
                blockedInfo = val.deviceInfo;
                deviceInfo = blockedInfo.deviceInfo;
                _deviceId4 = deviceInfo.deviceId;
                message = Object.assign({}, payload);
                message.code = blockedInfo.code;
                message.reason = blockedInfo.reason;

                if (message.code === "m.no_olm") {
                  delete message.room_id;
                  delete message.session_id;
                }

                if (!contentMap[_userId5]) {
                  contentMap[_userId5] = {};
                }

                contentMap[_userId5][_deviceId4] = message;
              }
            } catch (err) {
              _iterator4.e(err);
            } finally {
              _iterator4.f();
            }

            _context8.next = 5;
            return this._baseApis.sendToDevice("org.matrix.room_key.withheld", contentMap);

          case 5:
            // store that we successfully uploaded the keys of the current slice
            for (_i12 = 0, _Object$keys5 = Object.keys(contentMap); _i12 < _Object$keys5.length; _i12++) {
              userId = _Object$keys5[_i12];

              for (_i13 = 0, _Object$keys6 = Object.keys(contentMap[userId]); _i13 < _Object$keys6.length; _i13++) {
                deviceId = _Object$keys6[_i13];
                session.markNotifiedBlockedDevice(userId, deviceId);
              }
            }

          case 6:
          case "end":
            return _context8.stop();
        }
      }
    }, _callee8, this);
  }));

  return function (_x7, _x8, _x9) {
    return _ref8.apply(this, arguments);
  };
}();
/**
 * Re-shares a megolm session key with devices if the key has already been
 * sent to them.
 *
 * @param {string} senderKey The key of the originating device for the session
 * @param {string} sessionId ID of the outbound session to share
 * @param {string} userId ID of the user who owns the target device
 * @param {module:crypto/deviceinfo} device The target device
 */


MegolmEncryption.prototype.reshareKeyWithDevice = /*#__PURE__*/function () {
  var _ref9 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee9(senderKey, sessionId, userId, device) {
    var obSessionInfo, sentChainIndex, key, payload, encryptedContent;
    return _regenerator["default"].wrap(function _callee9$(_context9) {
      while (1) {
        switch (_context9.prev = _context9.next) {
          case 0:
            obSessionInfo = this._outboundSessions[sessionId];

            if (obSessionInfo) {
              _context9.next = 4;
              break;
            }

            _logger.logger.debug("megolm session ".concat(sessionId, " not found: not re-sharing keys"));

            return _context9.abrupt("return");

          case 4:
            if (!(obSessionInfo.sharedWithDevices[userId] === undefined)) {
              _context9.next = 7;
              break;
            }

            _logger.logger.debug("megolm session ".concat(sessionId, " never shared with user ").concat(userId));

            return _context9.abrupt("return");

          case 7:
            sentChainIndex = obSessionInfo.sharedWithDevices[userId][device.deviceId];

            if (!(sentChainIndex === undefined)) {
              _context9.next = 11;
              break;
            }

            _logger.logger.debug("megolm session ID " + sessionId + " never shared with device " + userId + ":" + device.deviceId);

            return _context9.abrupt("return");

          case 11:
            _context9.next = 13;
            return this._olmDevice.getInboundGroupSessionKey(this._roomId, senderKey, sessionId, sentChainIndex);

          case 13:
            key = _context9.sent;

            if (key) {
              _context9.next = 17;
              break;
            }

            _logger.logger.warn("No inbound session key found for megolm ".concat(sessionId, ": not re-sharing keys"));

            return _context9.abrupt("return");

          case 17:
            _context9.next = 19;
            return olmlib.ensureOlmSessionsForDevices(this._olmDevice, this._baseApis, (0, _defineProperty2["default"])({}, userId, [device]));

          case 19:
            payload = {
              type: "m.forwarded_room_key",
              content: {
                "algorithm": olmlib.MEGOLM_ALGORITHM,
                "room_id": this._roomId,
                "session_id": sessionId,
                "session_key": key.key,
                "chain_index": key.chain_index,
                "sender_key": senderKey,
                "sender_claimed_ed25519_key": key.sender_claimed_ed25519_key,
                "forwarding_curve25519_key_chain": key.forwarding_curve25519_key_chain,
                "org.matrix.msc3061.shared_history": key.shared_history || false
              }
            };
            encryptedContent = {
              algorithm: olmlib.OLM_ALGORITHM,
              sender_key: this._olmDevice.deviceCurve25519Key,
              ciphertext: {}
            };
            _context9.next = 23;
            return olmlib.encryptMessageForDevice(encryptedContent.ciphertext, this._userId, this._deviceId, this._olmDevice, userId, device, payload);

          case 23:
            _context9.next = 25;
            return this._baseApis.sendToDevice("m.room.encrypted", (0, _defineProperty2["default"])({}, userId, (0, _defineProperty2["default"])({}, device.deviceId, encryptedContent)));

          case 25:
            _logger.logger.debug("Re-shared key for megolm session ".concat(sessionId, " ") + "with ".concat(userId, ":").concat(device.deviceId));

          case 26:
          case "end":
            return _context9.stop();
        }
      }
    }, _callee9, this);
  }));

  return function (_x10, _x11, _x12, _x13) {
    return _ref9.apply(this, arguments);
  };
}();
/**
 * @private
 *
 * @param {module:crypto/algorithms/megolm.OutboundSessionInfo} session
 *
 * @param {object} key the session key as returned by
 *    OlmDevice.getOutboundGroupSessionKey
 *
 * @param {object} payload the base to-device message payload for sharing keys
 *
 * @param {object<string, module:crypto/deviceinfo[]>} devicesByUser
 *    map from userid to list of devices
 *
 * @param {array<object>} errorDevices
 *    array that will be populated with the devices that we can't get an
 *    olm session for
 *
 * @param {Number} [otkTimeout] The timeout in milliseconds when requesting
 *     one-time keys for establishing new olm sessions.
 *
 * @param {Array} [failedServers] An array to fill with remote servers that
 *     failed to respond to one-time-key requests.
 */


MegolmEncryption.prototype._shareKeyWithDevices = /*#__PURE__*/function () {
  var _ref10 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee10(session, key, payload, devicesByUser, errorDevices, otkTimeout, failedServers) {
    var devicemap;
    return _regenerator["default"].wrap(function _callee10$(_context10) {
      while (1) {
        switch (_context10.prev = _context10.next) {
          case 0:
            _logger.logger.debug("Ensuring Olm sessions for devices in ".concat(this._roomId));

            _context10.next = 3;
            return olmlib.ensureOlmSessionsForDevices(this._olmDevice, this._baseApis, devicesByUser, otkTimeout, failedServers, _logger.logger.withPrefix("[".concat(this._roomId, "]")));

          case 3:
            devicemap = _context10.sent;

            _logger.logger.debug("Ensured Olm sessions for devices in ".concat(this._roomId));

            this._getDevicesWithoutSessions(devicemap, devicesByUser, errorDevices);

            _logger.logger.debug("Sharing keys with Olm sessions in ".concat(this._roomId));

            _context10.next = 9;
            return this._shareKeyWithOlmSessions(session, key, payload, devicemap);

          case 9:
            _logger.logger.debug("Shared keys with Olm sessions in ".concat(this._roomId));

          case 10:
          case "end":
            return _context10.stop();
        }
      }
    }, _callee10, this);
  }));

  return function (_x14, _x15, _x16, _x17, _x18, _x19, _x20) {
    return _ref10.apply(this, arguments);
  };
}();

MegolmEncryption.prototype._shareKeyWithOlmSessions = /*#__PURE__*/function () {
  var _ref11 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee11(session, key, payload, devicemap) {
    var userDeviceMaps, i, taskDetail;
    return _regenerator["default"].wrap(function _callee11$(_context11) {
      while (1) {
        switch (_context11.prev = _context11.next) {
          case 0:
            userDeviceMaps = this._splitDevices(devicemap);
            i = 0;

          case 2:
            if (!(i < userDeviceMaps.length)) {
              _context11.next = 18;
              break;
            }

            taskDetail = "megolm keys for ".concat(session.sessionId, " ") + "in ".concat(this._roomId, " (slice ").concat(i + 1, "/").concat(userDeviceMaps.length, ")");
            _context11.prev = 4;

            _logger.logger.debug("Sharing ".concat(taskDetail));

            _context11.next = 8;
            return this._encryptAndSendKeysToDevices(session, key.chain_index, userDeviceMaps[i], payload);

          case 8:
            _logger.logger.debug("Shared ".concat(taskDetail));

            _context11.next = 15;
            break;

          case 11:
            _context11.prev = 11;
            _context11.t0 = _context11["catch"](4);

            _logger.logger.error("Failed to share ".concat(taskDetail));

            throw _context11.t0;

          case 15:
            i++;
            _context11.next = 2;
            break;

          case 18:
          case "end":
            return _context11.stop();
        }
      }
    }, _callee11, this, [[4, 11]]);
  }));

  return function (_x21, _x22, _x23, _x24) {
    return _ref11.apply(this, arguments);
  };
}();
/**
 * Notify devices that we weren't able to create olm sessions.
 *
 * @param {module:crypto/algorithms/megolm.OutboundSessionInfo} session
 *
 * @param {object} key
 *
 * @param {Array<object>} failedDevices the devices that we were unable to
 *     create olm sessions for, as returned by _shareKeyWithDevices
 */


MegolmEncryption.prototype._notifyFailedOlmDevices = /*#__PURE__*/function () {
  var _ref12 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee12(session, key, failedDevices) {
    var _iterator5, _step5, _step5$value, userId, deviceInfo, deviceId, filteredFailedDevices, blockedMap, _iterator6, _step6, _step6$value, _userId7, _deviceInfo2;

    return _regenerator["default"].wrap(function _callee12$(_context12) {
      while (1) {
        switch (_context12.prev = _context12.next) {
          case 0:
            _logger.logger.debug("Notifying ".concat(failedDevices.length, " devices we failed to ") + "create Olm sessions in ".concat(this._roomId)); // mark the devices that failed as "handled" because we don't want to try
            // to claim a one-time-key for dead devices on every message.


            _iterator5 = _createForOfIteratorHelper(failedDevices);

            try {
              for (_iterator5.s(); !(_step5 = _iterator5.n()).done;) {
                _step5$value = _step5.value, userId = _step5$value.userId, deviceInfo = _step5$value.deviceInfo;
                deviceId = deviceInfo.deviceId;
                session.markSharedWithDevice(userId, deviceId, key.chain_index);
              }
            } catch (err) {
              _iterator5.e(err);
            } finally {
              _iterator5.f();
            }

            _context12.next = 5;
            return this._olmDevice.filterOutNotifiedErrorDevices(failedDevices);

          case 5:
            filteredFailedDevices = _context12.sent;

            _logger.logger.debug("Filtered down to ".concat(filteredFailedDevices.length, " error devices ") + "in ".concat(this._roomId));

            blockedMap = {};
            _iterator6 = _createForOfIteratorHelper(filteredFailedDevices);

            try {
              for (_iterator6.s(); !(_step6 = _iterator6.n()).done;) {
                _step6$value = _step6.value, _userId7 = _step6$value.userId, _deviceInfo2 = _step6$value.deviceInfo;
                blockedMap[_userId7] = blockedMap[_userId7] || {}; // we use a similar format to what
                // olmlib.ensureOlmSessionsForDevices returns, so that
                // we can use the same function to split

                blockedMap[_userId7][_deviceInfo2.deviceId] = {
                  device: {
                    code: "m.no_olm",
                    reason: _OlmDevice.WITHHELD_MESSAGES["m.no_olm"],
                    deviceInfo: _deviceInfo2
                  }
                };
              } // send the notifications

            } catch (err) {
              _iterator6.e(err);
            } finally {
              _iterator6.f();
            }

            _context12.next = 12;
            return this._notifyBlockedDevices(session, blockedMap);

          case 12:
            _logger.logger.debug("Notified ".concat(filteredFailedDevices.length, " devices we failed to ") + "create Olm sessions in ".concat(this._roomId));

          case 13:
          case "end":
            return _context12.stop();
        }
      }
    }, _callee12, this);
  }));

  return function (_x25, _x26, _x27) {
    return _ref12.apply(this, arguments);
  };
}();
/**
 * Notify blocked devices that they have been blocked.
 *
 * @param {module:crypto/algorithms/megolm.OutboundSessionInfo} session
 *
 * @param {object<string, object>} devicesByUser
 *    map from userid to device ID to blocked data
 */


MegolmEncryption.prototype._notifyBlockedDevices = /*#__PURE__*/function () {
  var _ref13 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee13(session, devicesByUser) {
    var payload, userDeviceMaps, i;
    return _regenerator["default"].wrap(function _callee13$(_context13) {
      while (1) {
        switch (_context13.prev = _context13.next) {
          case 0:
            payload = {
              room_id: this._roomId,
              session_id: session.sessionId,
              algorithm: olmlib.MEGOLM_ALGORITHM,
              sender_key: this._olmDevice.deviceCurve25519Key
            };
            userDeviceMaps = this._splitDevices(devicesByUser);
            i = 0;

          case 3:
            if (!(i < userDeviceMaps.length)) {
              _context13.next = 17;
              break;
            }

            _context13.prev = 4;
            _context13.next = 7;
            return this._sendBlockedNotificationsToDevices(session, userDeviceMaps[i], payload);

          case 7:
            _logger.logger.log("Completed blacklist notification for ".concat(session.sessionId, " ") + "in ".concat(this._roomId, " (slice ").concat(i + 1, "/").concat(userDeviceMaps.length, ")"));

            _context13.next = 14;
            break;

          case 10:
            _context13.prev = 10;
            _context13.t0 = _context13["catch"](4);

            _logger.logger.log("blacklist notification for ".concat(session.sessionId, " in ") + "".concat(this._roomId, " (slice ").concat(i + 1, "/").concat(userDeviceMaps.length, ") failed"));

            throw _context13.t0;

          case 14:
            i++;
            _context13.next = 3;
            break;

          case 17:
          case "end":
            return _context13.stop();
        }
      }
    }, _callee13, this, [[4, 10]]);
  }));

  return function (_x28, _x29) {
    return _ref13.apply(this, arguments);
  };
}();
/**
 * Perform any background tasks that can be done before a message is ready to
 * send, in order to speed up sending of the message.
 *
 * @param {module:models/room} room the room the event is in
 */


MegolmEncryption.prototype.prepareToEncrypt = function (room) {
  var _this3 = this;

  if (this.encryptionPreparation) {
    // We're already preparing something, so don't do anything else.
    // FIXME: check if we need to restart
    // (https://github.com/matrix-org/matrix-js-sdk/issues/1255)
    var elapsedTime = Date.now() - this.encryptionPreparationMetadata.startTime;

    _logger.logger.debug("Already started preparing to encrypt for ".concat(this._roomId, " ") + "".concat(elapsedTime, " ms ago, skipping"));

    return;
  }

  _logger.logger.debug("Preparing to encrypt events for ".concat(this._roomId));

  this.encryptionPreparationMetadata = {
    startTime: Date.now()
  };
  this.encryptionPreparation = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee14() {
    var _yield$_this3$_getDev, _yield$_this3$_getDev2, devicesInRoom, blocked;

    return _regenerator["default"].wrap(function _callee14$(_context14) {
      while (1) {
        switch (_context14.prev = _context14.next) {
          case 0:
            _context14.prev = 0;

            _logger.logger.debug("Getting devices in ".concat(_this3._roomId));

            _context14.next = 4;
            return _this3._getDevicesInRoom(room);

          case 4:
            _yield$_this3$_getDev = _context14.sent;
            _yield$_this3$_getDev2 = (0, _slicedToArray2["default"])(_yield$_this3$_getDev, 2);
            devicesInRoom = _yield$_this3$_getDev2[0];
            blocked = _yield$_this3$_getDev2[1];

            if (_this3._crypto.getGlobalErrorOnUnknownDevices()) {
              // Drop unknown devices for now.  When the message gets sent, we'll
              // throw an error, but we'll still be prepared to send to the known
              // devices.
              _this3._removeUnknownDevices(devicesInRoom);
            }

            _logger.logger.debug("Ensuring outbound session in ".concat(_this3._roomId));

            _context14.next = 12;
            return _this3._ensureOutboundSession(room, devicesInRoom, blocked, true);

          case 12:
            _logger.logger.debug("Ready to encrypt events for ".concat(_this3._roomId));

            _context14.next = 18;
            break;

          case 15:
            _context14.prev = 15;
            _context14.t0 = _context14["catch"](0);

            _logger.logger.error("Failed to prepare to encrypt events for ".concat(_this3._roomId), _context14.t0);

          case 18:
            _context14.prev = 18;
            delete _this3.encryptionPreparationMetadata;
            delete _this3.encryptionPreparation;
            return _context14.finish(18);

          case 22:
          case "end":
            return _context14.stop();
        }
      }
    }, _callee14, null, [[0, 15, 18, 22]]);
  }))();
};
/**
 * @inheritdoc
 *
 * @param {module:models/room} room
 * @param {string} eventType
 * @param {object} content plaintext event content
 *
 * @return {Promise} Promise which resolves to the new event body
 */


MegolmEncryption.prototype.encryptMessage = /*#__PURE__*/function () {
  var _ref15 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee15(room, eventType, content) {
    var _yield$this$_getDevic, _yield$this$_getDevic2, devicesInRoom, blocked, session, payloadJson, ciphertext, encryptedContent;

    return _regenerator["default"].wrap(function _callee15$(_context15) {
      while (1) {
        switch (_context15.prev = _context15.next) {
          case 0:
            _logger.logger.log("Starting to encrypt event for ".concat(this._roomId));

            if (!this.encryptionPreparation) {
              _context15.next = 9;
              break;
            }

            _context15.prev = 2;
            _context15.next = 5;
            return this.encryptionPreparation;

          case 5:
            _context15.next = 9;
            break;

          case 7:
            _context15.prev = 7;
            _context15.t0 = _context15["catch"](2);

          case 9:
            _context15.next = 11;
            return this._getDevicesInRoom(room);

          case 11:
            _yield$this$_getDevic = _context15.sent;
            _yield$this$_getDevic2 = (0, _slicedToArray2["default"])(_yield$this$_getDevic, 2);
            devicesInRoom = _yield$this$_getDevic2[0];
            blocked = _yield$this$_getDevic2[1];

            // check if any of these devices are not yet known to the user.
            // if so, warn the user so they can verify or ignore.
            if (this._crypto.getGlobalErrorOnUnknownDevices()) {
              this._checkForUnknownDevices(devicesInRoom);
            }

            _context15.next = 18;
            return this._ensureOutboundSession(room, devicesInRoom, blocked);

          case 18:
            session = _context15.sent;
            payloadJson = {
              room_id: this._roomId,
              type: eventType,
              content: content
            };
            ciphertext = this._olmDevice.encryptGroupMessage(session.sessionId, JSON.stringify(payloadJson));
            encryptedContent = {
              algorithm: olmlib.MEGOLM_ALGORITHM,
              sender_key: this._olmDevice.deviceCurve25519Key,
              ciphertext: ciphertext,
              session_id: session.sessionId,
              // Include our device ID so that recipients can send us a
              // m.new_device message if they don't have our session key.
              // XXX: Do we still need this now that m.new_device messages
              // no longer exist since #483?
              device_id: this._deviceId
            };
            session.useCount++;
            return _context15.abrupt("return", encryptedContent);

          case 24:
          case "end":
            return _context15.stop();
        }
      }
    }, _callee15, this, [[2, 7]]);
  }));

  return function (_x30, _x31, _x32) {
    return _ref15.apply(this, arguments);
  };
}();
/**
 * Forces the current outbound group session to be discarded such
 * that another one will be created next time an event is sent.
 *
 * This should not normally be necessary.
 */


MegolmEncryption.prototype.forceDiscardSession = function () {
  this._setupPromise = this._setupPromise.then(function () {
    return null;
  });
};
/**
 * Checks the devices we're about to send to and see if any are entirely
 * unknown to the user.  If so, warn the user, and mark them as known to
 * give the user a chance to go verify them before re-sending this message.
 *
 * @param {Object} devicesInRoom userId -> {deviceId -> object}
 *   devices we should shared the session with.
 */


MegolmEncryption.prototype._checkForUnknownDevices = function (devicesInRoom) {
  var unknownDevices = {};
  Object.keys(devicesInRoom).forEach(function (userId) {
    Object.keys(devicesInRoom[userId]).forEach(function (deviceId) {
      var device = devicesInRoom[userId][deviceId];

      if (device.isUnverified() && !device.isKnown()) {
        if (!unknownDevices[userId]) {
          unknownDevices[userId] = {};
        }

        unknownDevices[userId][deviceId] = device;
      }
    });
  });

  if (Object.keys(unknownDevices).length) {
    // it'd be kind to pass unknownDevices up to the user in this error
    throw new _base.UnknownDeviceError("This room contains unknown devices which have not been verified. " + "We strongly recommend you verify them before continuing.", unknownDevices);
  }
};
/**
 * Remove unknown devices from a set of devices.  The devicesInRoom parameter
 * will be modified.
 *
 * @param {Object} devicesInRoom userId -> {deviceId -> object}
 *   devices we should shared the session with.
 */


MegolmEncryption.prototype._removeUnknownDevices = function (devicesInRoom) {
  for (var _i14 = 0, _Object$entries7 = Object.entries(devicesInRoom); _i14 < _Object$entries7.length; _i14++) {
    var _Object$entries7$_i = (0, _slicedToArray2["default"])(_Object$entries7[_i14], 2),
        userId = _Object$entries7$_i[0],
        userDevices = _Object$entries7$_i[1];

    for (var _i15 = 0, _Object$entries8 = Object.entries(userDevices); _i15 < _Object$entries8.length; _i15++) {
      var _Object$entries8$_i = (0, _slicedToArray2["default"])(_Object$entries8[_i15], 2),
          deviceId = _Object$entries8$_i[0],
          device = _Object$entries8$_i[1];

      if (device.isUnverified() && !device.isKnown()) {
        delete userDevices[deviceId];
      }
    }

    if (Object.keys(userDevices).length === 0) {
      delete devicesInRoom[userId];
    }
  }
};
/**
 * Get the list of unblocked devices for all users in the room
 *
 * @param {module:models/room} room
 *
 * @return {Promise} Promise which resolves to an array whose
 *     first element is a map from userId to deviceId to deviceInfo indicating
 *     the devices that messages should be encrypted to, and whose second
 *     element is a map from userId to deviceId to data indicating the devices
 *     that are in the room but that have been blocked
 */


MegolmEncryption.prototype._getDevicesInRoom = /*#__PURE__*/function () {
  var _ref16 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee16(room) {
    var members, roomMembers, isBlacklisting, devices, blocked, userId, userDevices, deviceId, deviceTrust, blockedInfo;
    return _regenerator["default"].wrap(function _callee16$(_context16) {
      while (1) {
        switch (_context16.prev = _context16.next) {
          case 0:
            _context16.next = 2;
            return room.getEncryptionTargetMembers();

          case 2:
            members = _context16.sent;
            roomMembers = utils.map(members, function (u) {
              return u.userId;
            }); // The global value is treated as a default for when rooms don't specify a value.

            isBlacklisting = this._crypto.getGlobalBlacklistUnverifiedDevices();

            if (typeof room.getBlacklistUnverifiedDevices() === 'boolean') {
              isBlacklisting = room.getBlacklistUnverifiedDevices();
            } // We are happy to use a cached version here: we assume that if we already
            // have a list of the user's devices, then we already share an e2e room
            // with them, which means that they will have announced any new devices via
            // device_lists in their /sync response.  This cache should then be maintained
            // using all the device_lists changes and left fields.
            // See https://github.com/vector-im/element-web/issues/2305 for details.


            _context16.next = 8;
            return this._crypto.downloadKeys(roomMembers, false);

          case 8:
            devices = _context16.sent;
            blocked = {}; // remove any blocked devices

            _context16.t0 = _regenerator["default"].keys(devices);

          case 11:
            if ((_context16.t1 = _context16.t0()).done) {
              _context16.next = 27;
              break;
            }

            userId = _context16.t1.value;

            if (devices.hasOwnProperty(userId)) {
              _context16.next = 15;
              break;
            }

            return _context16.abrupt("continue", 11);

          case 15:
            userDevices = devices[userId];
            _context16.t2 = _regenerator["default"].keys(userDevices);

          case 17:
            if ((_context16.t3 = _context16.t2()).done) {
              _context16.next = 25;
              break;
            }

            deviceId = _context16.t3.value;

            if (userDevices.hasOwnProperty(deviceId)) {
              _context16.next = 21;
              break;
            }

            return _context16.abrupt("continue", 17);

          case 21:
            deviceTrust = this._crypto.checkDeviceTrust(userId, deviceId);

            if (userDevices[deviceId].isBlocked() || !deviceTrust.isVerified() && isBlacklisting) {
              if (!blocked[userId]) {
                blocked[userId] = {};
              }

              blockedInfo = userDevices[deviceId].isBlocked() ? {
                code: "m.blacklisted",
                reason: _OlmDevice.WITHHELD_MESSAGES["m.blacklisted"]
              } : {
                code: "m.unverified",
                reason: _OlmDevice.WITHHELD_MESSAGES["m.unverified"]
              };
              blockedInfo.deviceInfo = userDevices[deviceId];
              blocked[userId][deviceId] = blockedInfo;
              delete userDevices[deviceId];
            }

            _context16.next = 17;
            break;

          case 25:
            _context16.next = 11;
            break;

          case 27:
            return _context16.abrupt("return", [devices, blocked]);

          case 28:
          case "end":
            return _context16.stop();
        }
      }
    }, _callee16, this);
  }));

  return function (_x33) {
    return _ref16.apply(this, arguments);
  };
}();
/**
 * Megolm decryption implementation
 *
 * @constructor
 * @extends {module:crypto/algorithms/DecryptionAlgorithm}
 *
 * @param {object} params parameters, as per
 *     {@link module:crypto/algorithms/DecryptionAlgorithm}
 */


function MegolmDecryption(params) {
  (0, utils.polyfillSuper)(this, _base.DecryptionAlgorithm, params); // events which we couldn't decrypt due to unknown sessions / indexes: map from
  // senderKey|sessionId to Set of MatrixEvents

  this._pendingEvents = {}; // this gets stubbed out by the unit tests.

  this.olmlib = olmlib;
}

utils.inherits(MegolmDecryption, _base.DecryptionAlgorithm);
var PROBLEM_DESCRIPTIONS = {
  no_olm: "The sender was unable to establish a secure channel.",
  unknown: "The secure channel with the sender was corrupted."
};
/**
 * @inheritdoc
 *
 * @param {MatrixEvent} event
 *
 * returns a promise which resolves to a
 * {@link module:crypto~EventDecryptionResult} once we have finished
 * decrypting, or rejects with an `algorithms.DecryptionError` if there is a
 * problem decrypting the event.
 */

MegolmDecryption.prototype.decryptEvent = /*#__PURE__*/function () {
  var _ref17 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee17(event) {
    var content, res, errorCode, problem, problemDescription, payload;
    return _regenerator["default"].wrap(function _callee17$(_context17) {
      while (1) {
        switch (_context17.prev = _context17.next) {
          case 0:
            content = event.getWireContent();

            if (!(!content.sender_key || !content.session_id || !content.ciphertext)) {
              _context17.next = 3;
              break;
            }

            throw new _base.DecryptionError("MEGOLM_MISSING_FIELDS", "Missing fields in input");

          case 3:
            // we add the event to the pending list *before* we start decryption.
            //
            // then, if the key turns up while decryption is in progress (and
            // decryption fails), we will schedule a retry.
            // (fixes https://github.com/vector-im/element-web/issues/5001)
            this._addEventToPendingList(event);

            _context17.prev = 4;
            _context17.next = 7;
            return this._olmDevice.decryptGroupMessage(event.getRoomId(), content.sender_key, content.session_id, content.ciphertext, event.getId(), event.getTs());

          case 7:
            res = _context17.sent;
            _context17.next = 17;
            break;

          case 10:
            _context17.prev = 10;
            _context17.t0 = _context17["catch"](4);

            if (!(_context17.t0.name === "DecryptionError")) {
              _context17.next = 14;
              break;
            }

            throw _context17.t0;

          case 14:
            errorCode = "OLM_DECRYPT_GROUP_MESSAGE_ERROR";

            if (_context17.t0 && _context17.t0.message === 'OLM.UNKNOWN_MESSAGE_INDEX') {
              this._requestKeysForEvent(event);

              errorCode = 'OLM_UNKNOWN_MESSAGE_INDEX';
            }

            throw new _base.DecryptionError(errorCode, _context17.t0 ? _context17.t0.toString() : "Unknown Error: Error is undefined", {
              session: content.sender_key + '|' + content.session_id
            });

          case 17:
            if (!(res === null)) {
              _context17.next = 27;
              break;
            }

            // We've got a message for a session we don't have.
            //
            // (XXX: We might actually have received this key since we started
            // decrypting, in which case we'll have scheduled a retry, and this
            // request will be redundant. We could probably check to see if the
            // event is still in the pending list; if not, a retry will have been
            // scheduled, so we needn't send out the request here.)
            this._requestKeysForEvent(event); // See if there was a problem with the olm session at the time the
            // event was sent.  Use a fuzz factor of 2 minutes.


            _context17.next = 21;
            return this._olmDevice.sessionMayHaveProblems(content.sender_key, event.getTs() - 120000);

          case 21:
            problem = _context17.sent;

            if (!problem) {
              _context17.next = 26;
              break;
            }

            problemDescription = PROBLEM_DESCRIPTIONS[problem.type] || PROBLEM_DESCRIPTIONS.unknown;

            if (problem.fixed) {
              problemDescription += " Trying to create a new secure channel and re-requesting the keys.";
            }

            throw new _base.DecryptionError("MEGOLM_UNKNOWN_INBOUND_SESSION_ID", problemDescription, {
              session: content.sender_key + '|' + content.session_id
            });

          case 26:
            throw new _base.DecryptionError("MEGOLM_UNKNOWN_INBOUND_SESSION_ID", "The sender's device has not sent us the keys for this message.", {
              session: content.sender_key + '|' + content.session_id
            });

          case 27:
            // success. We can remove the event from the pending list, if that hasn't
            // already happened.
            this._removeEventFromPendingList(event);

            payload = JSON.parse(res.result); // belt-and-braces check that the room id matches that indicated by the HS
            // (this is somewhat redundant, since the megolm session is scoped to the
            // room, so neither the sender nor a MITM can lie about the room_id).

            if (!(payload.room_id !== event.getRoomId())) {
              _context17.next = 31;
              break;
            }

            throw new _base.DecryptionError("MEGOLM_BAD_ROOM", "Message intended for room " + payload.room_id);

          case 31:
            return _context17.abrupt("return", {
              clearEvent: payload,
              senderCurve25519Key: res.senderKey,
              claimedEd25519Key: res.keysClaimed.ed25519,
              forwardingCurve25519KeyChain: res.forwardingCurve25519KeyChain,
              untrusted: res.untrusted
            });

          case 32:
          case "end":
            return _context17.stop();
        }
      }
    }, _callee17, this, [[4, 10]]);
  }));

  return function (_x34) {
    return _ref17.apply(this, arguments);
  };
}();

MegolmDecryption.prototype._requestKeysForEvent = function (event) {
  var wireContent = event.getWireContent();
  var recipients = event.getKeyRequestRecipients(this._userId);

  this._crypto.requestRoomKey({
    room_id: event.getRoomId(),
    algorithm: wireContent.algorithm,
    sender_key: wireContent.sender_key,
    session_id: wireContent.session_id
  }, recipients);
};
/**
 * Add an event to the list of those awaiting their session keys.
 *
 * @private
 *
 * @param {module:models/event.MatrixEvent} event
 */


MegolmDecryption.prototype._addEventToPendingList = function (event) {
  var content = event.getWireContent();
  var senderKey = content.sender_key;
  var sessionId = content.session_id;

  if (!this._pendingEvents[senderKey]) {
    this._pendingEvents[senderKey] = new Map();
  }

  var senderPendingEvents = this._pendingEvents[senderKey];

  if (!senderPendingEvents.has(sessionId)) {
    senderPendingEvents.set(sessionId, new Set());
  }

  senderPendingEvents.get(sessionId).add(event);
};
/**
 * Remove an event from the list of those awaiting their session keys.
 *
 * @private
 *
 * @param {module:models/event.MatrixEvent} event
 */


MegolmDecryption.prototype._removeEventFromPendingList = function (event) {
  var content = event.getWireContent();
  var senderKey = content.sender_key;
  var sessionId = content.session_id;
  var senderPendingEvents = this._pendingEvents[senderKey];
  var pendingEvents = senderPendingEvents && senderPendingEvents.get(sessionId);

  if (!pendingEvents) {
    return;
  }

  pendingEvents["delete"](event);

  if (pendingEvents.size === 0) {
    senderPendingEvents["delete"](senderKey);
  }

  if (senderPendingEvents.size === 0) {
    delete this._pendingEvents[senderKey];
  }
};
/**
 * @inheritdoc
 *
 * @param {module:models/event.MatrixEvent} event key event
 */


MegolmDecryption.prototype.onRoomKeyEvent = function (event) {
  var _this4 = this;

  var content = event.getContent();
  var sessionId = content.session_id;
  var senderKey = event.getSenderKey();
  var forwardingKeyChain = [];
  var exportFormat = false;
  var keysClaimed;

  if (!content.room_id || !sessionId || !content.session_key) {
    _logger.logger.error("key event is missing fields");

    return;
  }

  if (!senderKey) {
    _logger.logger.error("key event has no sender key (not encrypted?)");

    return;
  }

  if (event.getType() == "m.forwarded_room_key") {
    exportFormat = true;
    forwardingKeyChain = content.forwarding_curve25519_key_chain;

    if (!utils.isArray(forwardingKeyChain)) {
      forwardingKeyChain = [];
    } // copy content before we modify it


    forwardingKeyChain = forwardingKeyChain.slice();
    forwardingKeyChain.push(senderKey);
    senderKey = content.sender_key;

    if (!senderKey) {
      _logger.logger.error("forwarded_room_key event is missing sender_key field");

      return;
    }

    var ed25519Key = content.sender_claimed_ed25519_key;

    if (!ed25519Key) {
      _logger.logger.error("forwarded_room_key_event is missing sender_claimed_ed25519_key field");

      return;
    }

    keysClaimed = {
      ed25519: ed25519Key
    };
  } else {
    keysClaimed = event.getKeysClaimed();
  }

  var extraSessionData = {};

  if (content["org.matrix.msc3061.shared_history"]) {
    extraSessionData.sharedHistory = true;
  }

  return this._olmDevice.addInboundGroupSession(content.room_id, senderKey, forwardingKeyChain, sessionId, content.session_key, keysClaimed, exportFormat, extraSessionData).then(function () {
    // have another go at decrypting events sent with this session.
    _this4._retryDecryption(senderKey, sessionId).then(function (success) {
      // cancel any outstanding room key requests for this session.
      // Only do this if we managed to decrypt every message in the
      // session, because if we didn't, we leave the other key
      // requests in the hopes that someone sends us a key that
      // includes an earlier index.
      if (success) {
        _this4._crypto.cancelRoomKeyRequest({
          algorithm: content.algorithm,
          room_id: content.room_id,
          session_id: content.session_id,
          sender_key: senderKey
        });
      }
    });
  }).then(function () {
    // don't wait for the keys to be backed up for the server
    _this4._crypto.backupGroupSession(content.room_id, senderKey, forwardingKeyChain, content.session_id, content.session_key, keysClaimed, exportFormat);
  })["catch"](function (e) {
    _logger.logger.error("Error handling m.room_key_event: ".concat(e));
  });
};
/**
 * @inheritdoc
 *
 * @param {module:models/event.MatrixEvent} event key event
 */


MegolmDecryption.prototype.onRoomKeyWithheldEvent = /*#__PURE__*/function () {
  var _ref18 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee18(event) {
    var content, senderKey, sender, device, encryptedContent;
    return _regenerator["default"].wrap(function _callee18$(_context18) {
      while (1) {
        switch (_context18.prev = _context18.next) {
          case 0:
            content = event.getContent();
            senderKey = content.sender_key;

            if (!(content.code === "m.no_olm")) {
              _context18.next = 36;
              break;
            }

            sender = event.getSender();

            _logger.logger.warn("".concat(sender, ":").concat(senderKey, " was unable to establish an olm session with us")); // if the sender says that they haven't been able to establish an olm
            // session, let's proactively establish one
            // Note: after we record that the olm session has had a problem, we
            // trigger retrying decryption for all the messages from the sender's
            // key, so that we can update the error message to indicate the olm
            // session problem.


            _context18.next = 7;
            return this._olmDevice.getSessionIdForDevice(senderKey);

          case 7:
            if (!_context18.sent) {
              _context18.next = 13;
              break;
            }

            // a session has already been established, so we don't need to
            // create a new one.
            _logger.logger.debug("New session already created.  Not creating a new one.");

            _context18.next = 11;
            return this._olmDevice.recordSessionProblem(senderKey, "no_olm", true);

          case 11:
            this.retryDecryptionFromSender(senderKey);
            return _context18.abrupt("return");

          case 13:
            device = this._crypto._deviceList.getDeviceByIdentityKey(content.algorithm, senderKey);

            if (device) {
              _context18.next = 24;
              break;
            }

            _context18.next = 17;
            return this._crypto.downloadKeys([sender], false);

          case 17:
            device = this._crypto._deviceList.getDeviceByIdentityKey(content.algorithm, senderKey);

            if (device) {
              _context18.next = 24;
              break;
            }

            _logger.logger.info("Couldn't find device for identity key " + senderKey + ": not establishing session");

            _context18.next = 22;
            return this._olmDevice.recordSessionProblem(senderKey, "no_olm", false);

          case 22:
            this.retryDecryptionFromSender(senderKey);
            return _context18.abrupt("return");

          case 24:
            _context18.next = 26;
            return olmlib.ensureOlmSessionsForDevices(this._olmDevice, this._baseApis, (0, _defineProperty2["default"])({}, sender, [device]), false);

          case 26:
            encryptedContent = {
              algorithm: olmlib.OLM_ALGORITHM,
              sender_key: this._olmDevice.deviceCurve25519Key,
              ciphertext: {}
            };
            _context18.next = 29;
            return olmlib.encryptMessageForDevice(encryptedContent.ciphertext, this._userId, this._deviceId, this._olmDevice, sender, device, {
              type: "m.dummy"
            });

          case 29:
            _context18.next = 31;
            return this._olmDevice.recordSessionProblem(senderKey, "no_olm", true);

          case 31:
            this.retryDecryptionFromSender(senderKey);
            _context18.next = 34;
            return this._baseApis.sendToDevice("m.room.encrypted", (0, _defineProperty2["default"])({}, sender, (0, _defineProperty2["default"])({}, device.deviceId, encryptedContent)));

          case 34:
            _context18.next = 38;
            break;

          case 36:
            _context18.next = 38;
            return this._olmDevice.addInboundGroupSessionWithheld(content.room_id, senderKey, content.session_id, content.code, content.reason);

          case 38:
          case "end":
            return _context18.stop();
        }
      }
    }, _callee18, this);
  }));

  return function (_x35) {
    return _ref18.apply(this, arguments);
  };
}();
/**
 * @inheritdoc
 */


MegolmDecryption.prototype.hasKeysForKeyRequest = function (keyRequest) {
  var body = keyRequest.requestBody;
  return this._olmDevice.hasInboundSessionKeys(body.room_id, body.sender_key, body.session_id // TODO: ratchet index
  );
};
/**
 * @inheritdoc
 */


MegolmDecryption.prototype.shareKeysWithDevice = function (keyRequest) {
  var _this5 = this;

  var userId = keyRequest.userId;
  var deviceId = keyRequest.deviceId;

  var deviceInfo = this._crypto.getStoredDevice(userId, deviceId);

  var body = keyRequest.requestBody;
  this.olmlib.ensureOlmSessionsForDevices(this._olmDevice, this._baseApis, (0, _defineProperty2["default"])({}, userId, [deviceInfo])).then(function (devicemap) {
    var olmSessionResult = devicemap[userId][deviceId];

    if (!olmSessionResult.sessionId) {
      // no session with this device, probably because there
      // were no one-time keys.
      //
      // ensureOlmSessionsForUsers has already done the logging,
      // so just skip it.
      return null;
    }

    _logger.logger.log("sharing keys for session " + body.sender_key + "|" + body.session_id + " with device " + userId + ":" + deviceId);

    return _this5._buildKeyForwardingMessage(body.room_id, body.sender_key, body.session_id);
  }).then(function (payload) {
    var encryptedContent = {
      algorithm: olmlib.OLM_ALGORITHM,
      sender_key: _this5._olmDevice.deviceCurve25519Key,
      ciphertext: {}
    };
    return _this5.olmlib.encryptMessageForDevice(encryptedContent.ciphertext, _this5._userId, _this5._deviceId, _this5._olmDevice, userId, deviceInfo, payload).then(function () {
      var contentMap = (0, _defineProperty2["default"])({}, userId, (0, _defineProperty2["default"])({}, deviceId, encryptedContent)); // TODO: retries

      return _this5._baseApis.sendToDevice("m.room.encrypted", contentMap);
    });
  });
};

MegolmDecryption.prototype._buildKeyForwardingMessage = /*#__PURE__*/function () {
  var _ref19 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee19(roomId, senderKey, sessionId) {
    var key;
    return _regenerator["default"].wrap(function _callee19$(_context19) {
      while (1) {
        switch (_context19.prev = _context19.next) {
          case 0:
            _context19.next = 2;
            return this._olmDevice.getInboundGroupSessionKey(roomId, senderKey, sessionId);

          case 2:
            key = _context19.sent;
            return _context19.abrupt("return", {
              type: "m.forwarded_room_key",
              content: {
                "algorithm": olmlib.MEGOLM_ALGORITHM,
                "room_id": roomId,
                "sender_key": senderKey,
                "sender_claimed_ed25519_key": key.sender_claimed_ed25519_key,
                "session_id": sessionId,
                "session_key": key.key,
                "chain_index": key.chain_index,
                "forwarding_curve25519_key_chain": key.forwarding_curve25519_key_chain,
                "org.matrix.msc3061.shared_history": key.shared_history || false
              }
            });

          case 4:
          case "end":
            return _context19.stop();
        }
      }
    }, _callee19, this);
  }));

  return function (_x36, _x37, _x38) {
    return _ref19.apply(this, arguments);
  };
}();
/**
 * @inheritdoc
 *
 * @param {module:crypto/OlmDevice.MegolmSessionData} session
 * @param {object} [opts={}] options for the import
 * @param {boolean} [opts.untrusted] whether the key should be considered as untrusted
 * @param {string} [opts.source] where the key came from
 */


MegolmDecryption.prototype.importRoomKey = function (session) {
  var _this6 = this;

  var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var extraSessionData = {};

  if (opts.untrusted) {
    extraSessionData.untrusted = true;
  }

  if (session["org.matrix.msc3061.shared_history"]) {
    extraSessionData.sharedHistory = true;
  }

  return this._olmDevice.addInboundGroupSession(session.room_id, session.sender_key, session.forwarding_curve25519_key_chain, session.session_id, session.session_key, session.sender_claimed_keys, true, extraSessionData).then(function () {
    if (opts.source !== "backup") {
      // don't wait for it to complete
      _this6._crypto.backupGroupSession(session.room_id, session.sender_key, session.forwarding_curve25519_key_chain, session.session_id, session.session_key, session.sender_claimed_keys, true)["catch"](function (e) {
        // This throws if the upload failed, but this is fine
        // since it will have written it to the db and will retry.
        _logger.logger.log("Failed to back up megolm session", e);
      });
    } // have another go at decrypting events sent with this session.


    _this6._retryDecryption(session.sender_key, session.session_id);
  });
};
/**
 * Have another go at decrypting events after we receive a key. Resolves once
 * decryption has been re-attempted on all events.
 *
 * @private
 * @param {String} senderKey
 * @param {String} sessionId
 *
 * @return {Boolean} whether all messages were successfully decrypted
 */


MegolmDecryption.prototype._retryDecryption = /*#__PURE__*/function () {
  var _ref20 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee21(senderKey, sessionId) {
    var _this7 = this;

    var senderPendingEvents, pending;
    return _regenerator["default"].wrap(function _callee21$(_context21) {
      while (1) {
        switch (_context21.prev = _context21.next) {
          case 0:
            senderPendingEvents = this._pendingEvents[senderKey];

            if (senderPendingEvents) {
              _context21.next = 3;
              break;
            }

            return _context21.abrupt("return", true);

          case 3:
            pending = senderPendingEvents.get(sessionId);

            if (pending) {
              _context21.next = 6;
              break;
            }

            return _context21.abrupt("return", true);

          case 6:
            _logger.logger.debug("Retrying decryption on events", (0, _toConsumableArray2["default"])(pending));

            _context21.next = 9;
            return Promise.all((0, _toConsumableArray2["default"])(pending).map( /*#__PURE__*/function () {
              var _ref21 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee20(ev) {
                return _regenerator["default"].wrap(function _callee20$(_context20) {
                  while (1) {
                    switch (_context20.prev = _context20.next) {
                      case 0:
                        _context20.prev = 0;
                        _context20.next = 3;
                        return ev.attemptDecryption(_this7._crypto, true);

                      case 3:
                        _context20.next = 7;
                        break;

                      case 5:
                        _context20.prev = 5;
                        _context20.t0 = _context20["catch"](0);

                      case 7:
                      case "end":
                        return _context20.stop();
                    }
                  }
                }, _callee20, null, [[0, 5]]);
              }));

              return function (_x41) {
                return _ref21.apply(this, arguments);
              };
            }()));

          case 9:
            return _context21.abrupt("return", !(this._pendingEvents[senderKey] || {})[sessionId]);

          case 10:
          case "end":
            return _context21.stop();
        }
      }
    }, _callee21, this);
  }));

  return function (_x39, _x40) {
    return _ref20.apply(this, arguments);
  };
}();

MegolmDecryption.prototype.retryDecryptionFromSender = /*#__PURE__*/function () {
  var _ref22 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee24(senderKey) {
    var _this8 = this;

    var senderPendingEvents;
    return _regenerator["default"].wrap(function _callee24$(_context24) {
      while (1) {
        switch (_context24.prev = _context24.next) {
          case 0:
            senderPendingEvents = this._pendingEvents[senderKey];

            if (senderPendingEvents) {
              _context24.next = 3;
              break;
            }

            return _context24.abrupt("return", true);

          case 3:
            delete this._pendingEvents[senderKey];
            _context24.next = 6;
            return Promise.all((0, _toConsumableArray2["default"])(senderPendingEvents).map( /*#__PURE__*/function () {
              var _ref24 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee23(_ref23) {
                var _ref25, _sessionId, pending;

                return _regenerator["default"].wrap(function _callee23$(_context23) {
                  while (1) {
                    switch (_context23.prev = _context23.next) {
                      case 0:
                        _ref25 = (0, _slicedToArray2["default"])(_ref23, 2), _sessionId = _ref25[0], pending = _ref25[1];
                        _context23.next = 3;
                        return Promise.all((0, _toConsumableArray2["default"])(pending).map( /*#__PURE__*/function () {
                          var _ref26 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee22(ev) {
                            return _regenerator["default"].wrap(function _callee22$(_context22) {
                              while (1) {
                                switch (_context22.prev = _context22.next) {
                                  case 0:
                                    _context22.prev = 0;
                                    _context22.next = 3;
                                    return ev.attemptDecryption(_this8._crypto);

                                  case 3:
                                    _context22.next = 7;
                                    break;

                                  case 5:
                                    _context22.prev = 5;
                                    _context22.t0 = _context22["catch"](0);

                                  case 7:
                                  case "end":
                                    return _context22.stop();
                                }
                              }
                            }, _callee22, null, [[0, 5]]);
                          }));

                          return function (_x44) {
                            return _ref26.apply(this, arguments);
                          };
                        }()));

                      case 3:
                      case "end":
                        return _context23.stop();
                    }
                  }
                }, _callee23);
              }));

              return function (_x43) {
                return _ref24.apply(this, arguments);
              };
            }()));

          case 6:
            return _context24.abrupt("return", !this._pendingEvents[senderKey]);

          case 7:
          case "end":
            return _context24.stop();
        }
      }
    }, _callee24, this);
  }));

  return function (_x42) {
    return _ref22.apply(this, arguments);
  };
}();

MegolmDecryption.prototype.sendSharedHistoryInboundSessions = /*#__PURE__*/function () {
  var _ref27 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee25(devicesByUser) {
    var sharedHistorySessions, _iterator7, _step7, _step7$value, senderKey, sessionId, payload, promises, contentMap, _i16, _Object$entries9, _Object$entries9$_i, userId, devices, _iterator8, _step8, deviceInfo, encryptedContent, _i17, _Object$keys7, _userId9, _i18, _Object$keys8, deviceId;

    return _regenerator["default"].wrap(function _callee25$(_context25) {
      while (1) {
        switch (_context25.prev = _context25.next) {
          case 0:
            _context25.next = 2;
            return olmlib.ensureOlmSessionsForDevices(this._olmDevice, this._baseApis, devicesByUser);

          case 2:
            _logger.logger.log("sendSharedHistoryInboundSessions to users", Object.keys(devicesByUser));

            _context25.next = 5;
            return this._olmDevice.getSharedHistoryInboundGroupSessions(this._roomId);

          case 5:
            sharedHistorySessions = _context25.sent;

            _logger.logger.log("shared-history sessions", sharedHistorySessions);

            _iterator7 = _createForOfIteratorHelper(sharedHistorySessions);
            _context25.prev = 8;

            _iterator7.s();

          case 10:
            if ((_step7 = _iterator7.n()).done) {
              _context25.next = 28;
              break;
            }

            _step7$value = (0, _slicedToArray2["default"])(_step7.value, 2), senderKey = _step7$value[0], sessionId = _step7$value[1];
            _context25.next = 14;
            return this._buildKeyForwardingMessage(this._roomId, senderKey, sessionId);

          case 14:
            payload = _context25.sent;
            promises = [];
            contentMap = {};

            for (_i16 = 0, _Object$entries9 = Object.entries(devicesByUser); _i16 < _Object$entries9.length; _i16++) {
              _Object$entries9$_i = (0, _slicedToArray2["default"])(_Object$entries9[_i16], 2), userId = _Object$entries9$_i[0], devices = _Object$entries9$_i[1];
              contentMap[userId] = {};
              _iterator8 = _createForOfIteratorHelper(devices);

              try {
                for (_iterator8.s(); !(_step8 = _iterator8.n()).done;) {
                  deviceInfo = _step8.value;
                  encryptedContent = {
                    algorithm: olmlib.OLM_ALGORITHM,
                    sender_key: this._olmDevice.deviceCurve25519Key,
                    ciphertext: {}
                  };
                  contentMap[userId][deviceInfo.deviceId] = encryptedContent;
                  promises.push(olmlib.encryptMessageForDevice(encryptedContent.ciphertext, this._userId, this._deviceId, this._olmDevice, userId, deviceInfo, payload));
                }
              } catch (err) {
                _iterator8.e(err);
              } finally {
                _iterator8.f();
              }
            }

            _context25.next = 20;
            return Promise.all(promises);

          case 20:
            // prune out any devices that encryptMessageForDevice could not encrypt for,
            // in which case it will have just not added anything to the ciphertext object.
            // There's no point sending messages to devices if we couldn't encrypt to them,
            // since that's effectively a blank message.
            for (_i17 = 0, _Object$keys7 = Object.keys(contentMap); _i17 < _Object$keys7.length; _i17++) {
              _userId9 = _Object$keys7[_i17];

              for (_i18 = 0, _Object$keys8 = Object.keys(contentMap[_userId9]); _i18 < _Object$keys8.length; _i18++) {
                deviceId = _Object$keys8[_i18];

                if (Object.keys(contentMap[_userId9][deviceId].ciphertext).length === 0) {
                  _logger.logger.log("No ciphertext for device " + _userId9 + ":" + deviceId + ": pruning");

                  delete contentMap[_userId9][deviceId];
                }
              } // No devices left for that user? Strip that too.


              if (Object.keys(contentMap[_userId9]).length === 0) {
                _logger.logger.log("Pruned all devices for user " + _userId9);

                delete contentMap[_userId9];
              }
            } // Is there anything left?


            if (!(Object.keys(contentMap).length === 0)) {
              _context25.next = 24;
              break;
            }

            _logger.logger.log("No users left to send to: aborting");

            return _context25.abrupt("return");

          case 24:
            _context25.next = 26;
            return this._baseApis.sendToDevice("m.room.encrypted", contentMap);

          case 26:
            _context25.next = 10;
            break;

          case 28:
            _context25.next = 33;
            break;

          case 30:
            _context25.prev = 30;
            _context25.t0 = _context25["catch"](8);

            _iterator7.e(_context25.t0);

          case 33:
            _context25.prev = 33;

            _iterator7.f();

            return _context25.finish(33);

          case 36:
          case "end":
            return _context25.stop();
        }
      }
    }, _callee25, this, [[8, 30, 33, 36]]);
  }));

  return function (_x45) {
    return _ref27.apply(this, arguments);
  };
}();

(0, _base.registerAlgorithm)(olmlib.MEGOLM_ALGORITHM, MegolmEncryption, MegolmDecryption);