"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

var _typeof = require("@babel/runtime/helpers/typeof");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.OLM_ALGORITHM = exports.MEGOLM_BACKUP_ALGORITHM = exports.MEGOLM_ALGORITHM = void 0;
exports.decodeBase64 = decodeBase64;
exports.encodeBase64 = encodeBase64;
exports.encodeUnpaddedBase64 = encodeUnpaddedBase64;
exports.encryptMessageForDevice = encryptMessageForDevice;
exports.ensureOlmSessionsForDevices = ensureOlmSessionsForDevices;
exports.getExistingOlmSessions = getExistingOlmSessions;
exports.pkSign = pkSign;
exports.pkVerify = pkVerify;
exports.verifySignature = verifySignature;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _toConsumableArray2 = _interopRequireDefault(require("@babel/runtime/helpers/toConsumableArray"));

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _logger = require("../logger");

var utils = _interopRequireWildcard(require("../utils"));

var _anotherJson = _interopRequireDefault(require("another-json"));

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || _typeof(obj) !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _createForOfIteratorHelper(o, allowArrayLike) { var it = typeof Symbol !== "undefined" && o[Symbol.iterator] || o["@@iterator"]; if (!it) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = it.call(o); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

/**
 * matrix algorithm tag for olm
 */
var OLM_ALGORITHM = "m.olm.v1.curve25519-aes-sha2";
/**
 * matrix algorithm tag for megolm
 */

exports.OLM_ALGORITHM = OLM_ALGORITHM;
var MEGOLM_ALGORITHM = "m.megolm.v1.aes-sha2";
/**
 * matrix algorithm tag for megolm backups
 */

exports.MEGOLM_ALGORITHM = MEGOLM_ALGORITHM;
var MEGOLM_BACKUP_ALGORITHM = "m.megolm_backup.v1.curve25519-aes-sha2";
/**
 * Encrypt an event payload for an Olm device
 *
 * @param {Object<string, string>} resultsObject  The `ciphertext` property
 *   of the m.room.encrypted event to which to add our result
 *
 * @param {string} ourUserId
 * @param {string} ourDeviceId
 * @param {module:crypto/OlmDevice} olmDevice olm.js wrapper
 * @param {string} recipientUserId
 * @param {module:crypto/deviceinfo} recipientDevice
 * @param {object} payloadFields fields to include in the encrypted payload
 *
 * Returns a promise which resolves (to undefined) when the payload
 *    has been encrypted into `resultsObject`
 */

exports.MEGOLM_BACKUP_ALGORITHM = MEGOLM_BACKUP_ALGORITHM;

function encryptMessageForDevice(_x, _x2, _x3, _x4, _x5, _x6, _x7) {
  return _encryptMessageForDevice.apply(this, arguments);
}
/**
 * Get the existing olm sessions for the given devices, and the devices that
 * don't have olm sessions.
 *
 * @param {module:crypto/OlmDevice} olmDevice
 *
 * @param {module:base-apis~MatrixBaseApis} baseApis
 *
 * @param {object<string, module:crypto/deviceinfo[]>} devicesByUser
 *    map from userid to list of devices to ensure sessions for
 *
 * @return {Promise} resolves to an array.  The first element of the array is a
 *    a map of user IDs to arrays of deviceInfo, representing the devices that
 *    don't have established olm sessions.  The second element of the array is
 *    a map from userId to deviceId to {@link module:crypto~OlmSessionResult}
 */


function _encryptMessageForDevice() {
  _encryptMessageForDevice = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(resultsObject, ourUserId, ourDeviceId, olmDevice, recipientUserId, recipientDevice, payloadFields) {
    var deviceKey, sessionId, payload;
    return _regenerator["default"].wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            deviceKey = recipientDevice.getIdentityKey();
            _context.next = 3;
            return olmDevice.getSessionIdForDevice(deviceKey);

          case 3:
            sessionId = _context.sent;

            if (!(sessionId === null)) {
              _context.next = 6;
              break;
            }

            return _context.abrupt("return");

          case 6:
            _logger.logger.log("Using sessionid " + sessionId + " for device " + recipientUserId + ":" + recipientDevice.deviceId);

            payload = {
              sender: ourUserId,
              sender_device: ourDeviceId,
              // Include the Ed25519 key so that the recipient knows what
              // device this message came from.
              // We don't need to include the curve25519 key since the
              // recipient will already know this from the olm headers.
              // When combined with the device keys retrieved from the
              // homeserver signed by the ed25519 key this proves that
              // the curve25519 key and the ed25519 key are owned by
              // the same device.
              keys: {
                "ed25519": olmDevice.deviceEd25519Key
              },
              // include the recipient device details in the payload,
              // to avoid unknown key attacks, per
              // https://github.com/vector-im/vector-web/issues/2483
              recipient: recipientUserId,
              recipient_keys: {
                "ed25519": recipientDevice.getFingerprint()
              }
            }; // TODO: technically, a bunch of that stuff only needs to be included for
            // pre-key messages: after that, both sides know exactly which devices are
            // involved in the session. If we're looking to reduce data transfer in the
            // future, we could elide them for subsequent messages.

            utils.extend(payload, payloadFields);
            _context.next = 11;
            return olmDevice.encryptMessage(deviceKey, sessionId, JSON.stringify(payload));

          case 11:
            resultsObject[deviceKey] = _context.sent;

          case 12:
          case "end":
            return _context.stop();
        }
      }
    }, _callee);
  }));
  return _encryptMessageForDevice.apply(this, arguments);
}

function getExistingOlmSessions(_x8, _x9, _x10) {
  return _getExistingOlmSessions.apply(this, arguments);
}
/**
 * Try to make sure we have established olm sessions for the given devices.
 *
 * @param {module:crypto/OlmDevice} olmDevice
 *
 * @param {module:base-apis~MatrixBaseApis} baseApis
 *
 * @param {object<string, module:crypto/deviceinfo[]>} devicesByUser
 *    map from userid to list of devices to ensure sessions for
 *
 * @param {boolean} [force=false] If true, establish a new session even if one
 *     already exists.
 *
 * @param {Number} [otkTimeout] The timeout in milliseconds when requesting
 *     one-time keys for establishing new olm sessions.
 *
 * @param {Array} [failedServers] An array to fill with remote servers that
 *     failed to respond to one-time-key requests.
 *
 * @param {Logger} [log] A possibly customised log
 *
 * @return {Promise} resolves once the sessions are complete, to
 *    an Object mapping from userId to deviceId to
 *    {@link module:crypto~OlmSessionResult}
 */


function _getExistingOlmSessions() {
  _getExistingOlmSessions = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(olmDevice, baseApis, devicesByUser) {
    var devicesWithoutSession, sessions, promises, _loop, _i, _Object$entries;

    return _regenerator["default"].wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            devicesWithoutSession = {};
            sessions = {};
            promises = [];

            _loop = function _loop() {
              var _Object$entries$_i = (0, _slicedToArray2["default"])(_Object$entries[_i], 2),
                  userId = _Object$entries$_i[0],
                  devices = _Object$entries$_i[1];

              var _iterator = _createForOfIteratorHelper(devices),
                  _step;

              try {
                var _loop2 = function _loop2() {
                  var deviceInfo = _step.value;
                  var deviceId = deviceInfo.deviceId;
                  var key = deviceInfo.getIdentityKey();
                  promises.push((0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
                    var sessionId;
                    return _regenerator["default"].wrap(function _callee2$(_context2) {
                      while (1) {
                        switch (_context2.prev = _context2.next) {
                          case 0:
                            _context2.next = 2;
                            return olmDevice.getSessionIdForDevice(key, true);

                          case 2:
                            sessionId = _context2.sent;

                            if (sessionId === null) {
                              devicesWithoutSession[userId] = devicesWithoutSession[userId] || [];
                              devicesWithoutSession[userId].push(deviceInfo);
                            } else {
                              sessions[userId] = sessions[userId] || {};
                              sessions[userId][deviceId] = {
                                device: deviceInfo,
                                sessionId: sessionId
                              };
                            }

                          case 4:
                          case "end":
                            return _context2.stop();
                        }
                      }
                    }, _callee2);
                  }))());
                };

                for (_iterator.s(); !(_step = _iterator.n()).done;) {
                  _loop2();
                }
              } catch (err) {
                _iterator.e(err);
              } finally {
                _iterator.f();
              }
            };

            for (_i = 0, _Object$entries = Object.entries(devicesByUser); _i < _Object$entries.length; _i++) {
              _loop();
            }

            _context3.next = 7;
            return Promise.all(promises);

          case 7:
            return _context3.abrupt("return", [devicesWithoutSession, sessions]);

          case 8:
          case "end":
            return _context3.stop();
        }
      }
    }, _callee3);
  }));
  return _getExistingOlmSessions.apply(this, arguments);
}

function ensureOlmSessionsForDevices(_x11, _x12, _x13, _x14, _x15, _x16, _x17) {
  return _ensureOlmSessionsForDevices.apply(this, arguments);
}

function _ensureOlmSessionsForDevices() {
  _ensureOlmSessionsForDevices = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4(olmDevice, baseApis, devicesByUser, force, otkTimeout, failedServers, log) {
    var devicesWithoutSession, result, resolveSession, _i2, _Object$entries2, _Object$entries2$_i, devices, _iterator2, _step2, _loop3, _ret, _i3, _Object$entries3, _Object$entries3$_i, userId, _devices, _iterator3, _step3, deviceInfo, deviceId, key, forWhom, sessionId, oneTimeKeyAlgorithm, res, taskDetail, _i4, _Object$values, resolver, _failedServers, otkResult, promises, _loop4, _i5, _Object$entries4;

    return _regenerator["default"].wrap(function _callee4$(_context4) {
      while (1) {
        switch (_context4.prev = _context4.next) {
          case 0:
            if (typeof force === "number") {
              log = failedServers;
              failedServers = otkTimeout;
              otkTimeout = force;
              force = false;
            }

            if (!log) {
              log = _logger.logger;
            }

            devicesWithoutSession = [// [userId, deviceId], ...
            ];
            result = {};
            resolveSession = {}; // Mark all sessions this task intends to update as in progress. It is
            // important to do this for all devices this task cares about in a single
            // synchronous operation, as otherwise it is possible to have deadlocks
            // where multiple tasks wait indefinitely on another task to update some set
            // of common devices.

            _i2 = 0, _Object$entries2 = Object.entries(devicesByUser);

          case 6:
            if (!(_i2 < _Object$entries2.length)) {
              _context4.next = 29;
              break;
            }

            _Object$entries2$_i = (0, _slicedToArray2["default"])(_Object$entries2[_i2], 2), devices = _Object$entries2$_i[1];
            _iterator2 = _createForOfIteratorHelper(devices);
            _context4.prev = 9;

            _loop3 = function _loop3() {
              var deviceInfo = _step2.value;
              var key = deviceInfo.getIdentityKey();

              if (key === olmDevice.deviceCurve25519Key) {
                // We don't start sessions with ourself, so there's no need to
                // mark it in progress.
                return "continue";
              }

              if (!olmDevice._sessionsInProgress[key]) {
                // pre-emptively mark the session as in-progress to avoid race
                // conditions.  If we find that we already have a session, then
                // we'll resolve
                olmDevice._sessionsInProgress[key] = new Promise(function (resolve) {
                  resolveSession[key] = function () {
                    delete olmDevice._sessionsInProgress[key];
                    resolve.apply(void 0, arguments);
                  };
                });
              }
            };

            _iterator2.s();

          case 12:
            if ((_step2 = _iterator2.n()).done) {
              _context4.next = 18;
              break;
            }

            _ret = _loop3();

            if (!(_ret === "continue")) {
              _context4.next = 16;
              break;
            }

            return _context4.abrupt("continue", 16);

          case 16:
            _context4.next = 12;
            break;

          case 18:
            _context4.next = 23;
            break;

          case 20:
            _context4.prev = 20;
            _context4.t0 = _context4["catch"](9);

            _iterator2.e(_context4.t0);

          case 23:
            _context4.prev = 23;

            _iterator2.f();

            return _context4.finish(23);

          case 26:
            _i2++;
            _context4.next = 6;
            break;

          case 29:
            _i3 = 0, _Object$entries3 = Object.entries(devicesByUser);

          case 30:
            if (!(_i3 < _Object$entries3.length)) {
              _context4.next = 64;
              break;
            }

            _Object$entries3$_i = (0, _slicedToArray2["default"])(_Object$entries3[_i3], 2), userId = _Object$entries3$_i[0], _devices = _Object$entries3$_i[1];
            result[userId] = {};
            _iterator3 = _createForOfIteratorHelper(_devices);
            _context4.prev = 34;

            _iterator3.s();

          case 36:
            if ((_step3 = _iterator3.n()).done) {
              _context4.next = 53;
              break;
            }

            deviceInfo = _step3.value;
            deviceId = deviceInfo.deviceId;
            key = deviceInfo.getIdentityKey();

            if (!(key === olmDevice.deviceCurve25519Key)) {
              _context4.next = 44;
              break;
            }

            // We should never be trying to start a session with ourself.
            // Apart from talking to yourself being the first sign of madness,
            // olm sessions can't do this because they get confused when
            // they get a message and see that the 'other side' has started a
            // new chain when this side has an active sender chain.
            // If you see this message being logged in the wild, we should find
            // the thing that is trying to send Olm messages to itself and fix it.
            log.info("Attempted to start session with ourself! Ignoring"); // We must fill in the section in the return value though, as callers
            // expect it to be there.

            result[userId][deviceId] = {
              device: deviceInfo,
              sessionId: null
            };
            return _context4.abrupt("continue", 51);

          case 44:
            forWhom = "for ".concat(key, " (").concat(userId, ":").concat(deviceId, ")");
            _context4.next = 47;
            return olmDevice.getSessionIdForDevice(key, resolveSession[key], log);

          case 47:
            sessionId = _context4.sent;

            if (sessionId !== null && resolveSession[key]) {
              // we found a session, but we had marked the session as
              // in-progress, so resolve it now, which will unmark it and
              // unblock anything that was waiting
              resolveSession[key]();
            }

            if (sessionId === null || force) {
              if (force) {
                log.info("Forcing new Olm session ".concat(forWhom));
              } else {
                log.info("Making new Olm session ".concat(forWhom));
              }

              devicesWithoutSession.push([userId, deviceId]);
            }

            result[userId][deviceId] = {
              device: deviceInfo,
              sessionId: sessionId
            };

          case 51:
            _context4.next = 36;
            break;

          case 53:
            _context4.next = 58;
            break;

          case 55:
            _context4.prev = 55;
            _context4.t1 = _context4["catch"](34);

            _iterator3.e(_context4.t1);

          case 58:
            _context4.prev = 58;

            _iterator3.f();

            return _context4.finish(58);

          case 61:
            _i3++;
            _context4.next = 30;
            break;

          case 64:
            if (!(devicesWithoutSession.length === 0)) {
              _context4.next = 66;
              break;
            }

            return _context4.abrupt("return", result);

          case 66:
            oneTimeKeyAlgorithm = "signed_curve25519";
            taskDetail = "one-time keys for ".concat(devicesWithoutSession.length, " devices");
            _context4.prev = 68;
            log.debug("Claiming ".concat(taskDetail));
            _context4.next = 72;
            return baseApis.claimOneTimeKeys(devicesWithoutSession, oneTimeKeyAlgorithm, otkTimeout);

          case 72:
            res = _context4.sent;
            log.debug("Claimed ".concat(taskDetail));
            _context4.next = 81;
            break;

          case 76:
            _context4.prev = 76;
            _context4.t2 = _context4["catch"](68);

            for (_i4 = 0, _Object$values = Object.values(resolveSession); _i4 < _Object$values.length; _i4++) {
              resolver = _Object$values[_i4];
              resolver();
            }

            log.log("Failed to claim ".concat(taskDetail), _context4.t2, devicesWithoutSession);
            throw _context4.t2;

          case 81:
            if (failedServers && "failures" in res) {
              (_failedServers = failedServers).push.apply(_failedServers, (0, _toConsumableArray2["default"])(Object.keys(res.failures)));
            }

            otkResult = res.one_time_keys || {};
            promises = [];

            _loop4 = function _loop4() {
              var _Object$entries4$_i = (0, _slicedToArray2["default"])(_Object$entries4[_i5], 2),
                  userId = _Object$entries4$_i[0],
                  devices = _Object$entries4$_i[1];

              var userRes = otkResult[userId] || {};

              var _loop5 = function _loop5(j) {
                var deviceInfo = devices[j];
                var deviceId = deviceInfo.deviceId;
                var key = deviceInfo.getIdentityKey();

                if (key === olmDevice.deviceCurve25519Key) {
                  // We've already logged about this above. Skip here too
                  // otherwise we'll log saying there are no one-time keys
                  // which will be confusing.
                  return "continue";
                }

                if (result[userId][deviceId].sessionId && !force) {
                  // we already have a result for this device
                  return "continue";
                }

                var deviceRes = userRes[deviceId] || {};
                var oneTimeKey = null;

                for (var keyId in deviceRes) {
                  if (keyId.indexOf(oneTimeKeyAlgorithm + ":") === 0) {
                    oneTimeKey = deviceRes[keyId];
                  }
                }

                if (!oneTimeKey) {
                  log.warn("No one-time keys (alg=".concat(oneTimeKeyAlgorithm, ") ") + "for device ".concat(userId, ":").concat(deviceId));

                  if (resolveSession[key]) {
                    resolveSession[key]();
                  }

                  return "continue";
                }

                promises.push(_verifyKeyAndStartSession(olmDevice, oneTimeKey, userId, deviceInfo).then(function (sid) {
                  if (resolveSession[key]) {
                    resolveSession[key](sid);
                  }

                  result[userId][deviceId].sessionId = sid;
                }, function (e) {
                  if (resolveSession[key]) {
                    resolveSession[key]();
                  }

                  throw e;
                }));
              };

              for (var j = 0; j < devices.length; j++) {
                var _ret2 = _loop5(j);

                if (_ret2 === "continue") continue;
              }
            };

            for (_i5 = 0, _Object$entries4 = Object.entries(devicesByUser); _i5 < _Object$entries4.length; _i5++) {
              _loop4();
            }

            taskDetail = "Olm sessions for ".concat(promises.length, " devices");
            log.debug("Starting ".concat(taskDetail));
            _context4.next = 90;
            return Promise.all(promises);

          case 90:
            log.debug("Started ".concat(taskDetail));
            return _context4.abrupt("return", result);

          case 92:
          case "end":
            return _context4.stop();
        }
      }
    }, _callee4, null, [[9, 20, 23, 26], [34, 55, 58, 61], [68, 76]]);
  }));
  return _ensureOlmSessionsForDevices.apply(this, arguments);
}

function _verifyKeyAndStartSession(_x18, _x19, _x20, _x21) {
  return _verifyKeyAndStartSession2.apply(this, arguments);
}
/**
 * Verify the signature on an object
 *
 * @param {module:crypto/OlmDevice} olmDevice olm wrapper to use for verify op
 *
 * @param {Object} obj object to check signature on.
 *
 * @param {string} signingUserId  ID of the user whose signature should be checked
 *
 * @param {string} signingDeviceId  ID of the device whose signature should be checked
 *
 * @param {string} signingKey   base64-ed ed25519 public key
 *
 * Returns a promise which resolves (to undefined) if the the signature is good,
 * or rejects with an Error if it is bad.
 */


function _verifyKeyAndStartSession2() {
  _verifyKeyAndStartSession2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5(olmDevice, oneTimeKey, userId, deviceInfo) {
    var deviceId, sid;
    return _regenerator["default"].wrap(function _callee5$(_context5) {
      while (1) {
        switch (_context5.prev = _context5.next) {
          case 0:
            deviceId = deviceInfo.deviceId;
            _context5.prev = 1;
            _context5.next = 4;
            return verifySignature(olmDevice, oneTimeKey, userId, deviceId, deviceInfo.getFingerprint());

          case 4:
            _context5.next = 10;
            break;

          case 6:
            _context5.prev = 6;
            _context5.t0 = _context5["catch"](1);

            _logger.logger.error("Unable to verify signature on one-time key for device " + userId + ":" + deviceId + ":", _context5.t0);

            return _context5.abrupt("return", null);

          case 10:
            _context5.prev = 10;
            _context5.next = 13;
            return olmDevice.createOutboundSession(deviceInfo.getIdentityKey(), oneTimeKey.key);

          case 13:
            sid = _context5.sent;
            _context5.next = 20;
            break;

          case 16:
            _context5.prev = 16;
            _context5.t1 = _context5["catch"](10);

            // possibly a bad key
            _logger.logger.error("Error starting olm session with device " + userId + ":" + deviceId + ": " + _context5.t1);

            return _context5.abrupt("return", null);

          case 20:
            _logger.logger.log("Started new olm sessionid " + sid + " for device " + userId + ":" + deviceId);

            return _context5.abrupt("return", sid);

          case 22:
          case "end":
            return _context5.stop();
        }
      }
    }, _callee5, null, [[1, 6], [10, 16]]);
  }));
  return _verifyKeyAndStartSession2.apply(this, arguments);
}

function verifySignature(_x22, _x23, _x24, _x25, _x26) {
  return _verifySignature.apply(this, arguments);
}
/**
 * Sign a JSON object using public key cryptography
 * @param {Object} obj Object to sign.  The object will be modified to include
 *     the new signature
 * @param {Olm.PkSigning|Uint8Array} key the signing object or the private key
 * seed
 * @param {string} userId The user ID who owns the signing key
 * @param {string} pubkey The public key (ignored if key is a seed)
 * @returns {string} the signature for the object
 */


function _verifySignature() {
  _verifySignature = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6(olmDevice, obj, signingUserId, signingDeviceId, signingKey) {
    var signKeyId, signatures, userSigs, signature, mangledObj, json;
    return _regenerator["default"].wrap(function _callee6$(_context6) {
      while (1) {
        switch (_context6.prev = _context6.next) {
          case 0:
            signKeyId = "ed25519:" + signingDeviceId;
            signatures = obj.signatures || {};
            userSigs = signatures[signingUserId] || {};
            signature = userSigs[signKeyId];

            if (signature) {
              _context6.next = 6;
              break;
            }

            throw Error("No signature");

          case 6:
            // prepare the canonical json: remove unsigned and signatures, and stringify with
            // anotherjson
            mangledObj = Object.assign({}, obj);
            delete mangledObj.unsigned;
            delete mangledObj.signatures;
            json = _anotherJson["default"].stringify(mangledObj);
            olmDevice.verifySignature(signingKey, json, signature);

          case 11:
          case "end":
            return _context6.stop();
        }
      }
    }, _callee6);
  }));
  return _verifySignature.apply(this, arguments);
}

function pkSign(obj, key, userId, pubkey) {
  var createdKey = false;

  if (key instanceof Uint8Array) {
    var keyObj = new global.Olm.PkSigning();
    pubkey = keyObj.init_with_seed(key);
    key = keyObj;
    createdKey = true;
  }

  var sigs = obj.signatures || {};
  delete obj.signatures;
  var unsigned = obj.unsigned;
  if (obj.unsigned) delete obj.unsigned;

  try {
    var mysigs = sigs[userId] || {};
    sigs[userId] = mysigs;
    return mysigs['ed25519:' + pubkey] = key.sign(_anotherJson["default"].stringify(obj));
  } finally {
    obj.signatures = sigs;
    if (unsigned) obj.unsigned = unsigned;

    if (createdKey) {
      key.free();
    }
  }
}
/**
 * Verify a signed JSON object
 * @param {Object} obj Object to verify
 * @param {string} pubkey The public key to use to verify
 * @param {string} userId The user ID who signed the object
 */


function pkVerify(obj, pubkey, userId) {
  var keyId = "ed25519:" + pubkey;

  if (!(obj.signatures && obj.signatures[userId] && obj.signatures[userId][keyId])) {
    throw new Error("No signature");
  }

  var signature = obj.signatures[userId][keyId];
  var util = new global.Olm.Utility();
  var sigs = obj.signatures;
  delete obj.signatures;
  var unsigned = obj.unsigned;
  if (obj.unsigned) delete obj.unsigned;

  try {
    util.ed25519_verify(pubkey, _anotherJson["default"].stringify(obj), signature);
  } finally {
    obj.signatures = sigs;
    if (unsigned) obj.unsigned = unsigned;
    util.free();
  }
}
/**
 * Encode a typed array of uint8 as base64.
 * @param {Uint8Array} uint8Array The data to encode.
 * @return {string} The base64.
 */


function encodeBase64(uint8Array) {
  return Buffer.from(uint8Array).toString("base64");
}
/**
 * Encode a typed array of uint8 as unpadded base64.
 * @param {Uint8Array} uint8Array The data to encode.
 * @return {string} The unpadded base64.
 */


function encodeUnpaddedBase64(uint8Array) {
  return encodeBase64(uint8Array).replace(/=+$/g, '');
}
/**
 * Decode a base64 string to a typed array of uint8.
 * @param {string} base64 The base64 to decode.
 * @return {Uint8Array} The decoded data.
 */


function decodeBase64(base64) {
  return Buffer.from(base64, "base64");
}